/*
 * Copyright (C) 1997-2011 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file afh_common.c Common audio format handler functions. */

#include <sys/mman.h> /* mmap */
#include <sys/time.h> /* gettimeofday */
#include <sys/types.h>
#include <dirent.h>
#include <regex.h>

#include "para.h"
#include "error.h"
#include "string.h"
#include "afh.h"

/* The mp3 audio format handler does not need any libs. */
void mp3_init(struct audio_format_handler *);

#ifdef HAVE_OGGVORBIS
	void ogg_init(struct audio_format_handler *);
#endif
#ifdef HAVE_FAAD
	void aac_afh_init(struct audio_format_handler *);
#endif
#ifdef HAVE_SPEEX
	void spx_afh_init(struct audio_format_handler *);
#endif

void wma_afh_init(struct audio_format_handler *);
/**
 * The list of supported audio formats.
 *
 * We always define the full array of audio formats even if some audio formats
 * were not compiled in. This is because for each audio file the number of its
 * audio format is stored in the database. We don't want that numbers to become
 * stale just because the user installed a new version of paraslash that
 * supports a different set of audio formats.
 *
 * It can still be easily detected whether an audio format is compiled in by
 * checking if the init function pointer is not \p NULL.
 */
static struct audio_format_handler afl[] = {
	{
		.name = "mp3",
		.init = mp3_init,
	},
	{
		.name = "ogg",
#ifdef HAVE_OGGVORBIS
		.init = ogg_init,
#endif
	},
	{
		.name = "aac",
#ifdef HAVE_FAAD
		.init = aac_afh_init,
#endif
	},
	{
		.name = "wma",
		.init = wma_afh_init,
	},
	{
		.name = "spx",
#ifdef HAVE_SPEEX
		.init = spx_afh_init,
#endif
	},
	{
		.name = NULL,
	}
};

static inline int next_audio_format(int format)
{
	for (;;) {
		if (!afl[format].name)
			return format;
		format++;
		if (afl[format].init)
			return format;
	}

}

/** Iterate over each supported audio format. */
#define FOR_EACH_AUDIO_FORMAT(i) for (i = 0; afl[i].name; i = next_audio_format(i))

/**
 * Call the init function of each supported audio format handler.
 */
void afh_init(void)
{
	int i;

	PARA_INFO_LOG("supported audio formats: %s\n",
		SERVER_AUDIO_FORMATS);
	FOR_EACH_AUDIO_FORMAT(i) {
		PARA_NOTICE_LOG("initializing %s handler\n",
			audio_format_name(i));
		afl[i].init(&afl[i]);
	}
}

/**
 * Guess the audio format judging from filename.
 *
 * \param name The filename.
 *
 * \return This function returns \p -E_AUDIO_FORMAT if it has no idea what kind
 * of audio file this might be. Otherwise the (non-negative) number of the
 * audio format is returned.
 */
int guess_audio_format(const char *name)
{
	int i,j, len = strlen(name);

	FOR_EACH_AUDIO_FORMAT(i) {
		for (j = 0; afl[i].suffixes[j]; j++) {
			const char *p = afl[i].suffixes[j];
			int plen = strlen(p);
			if (len < plen + 1)
				continue;
			if (name[len - plen - 1] != '.')
				continue;
			if (strcasecmp(name + len - plen, p))
				continue;
//			PARA_DEBUG_LOG("might be %s\n", audio_format_name(i));
			return i;
		}
	}
	return -E_AUDIO_FORMAT;
}

/**
 * Call get_file_info() to obtain an afhi structure.
 *
 * \param path The full path of the audio file.
 * \param data Pointer to the contents of the (mapped) file.
 * \param size The file size in bytes.
 * \param fd The open file descriptor.
 * \param afhi Result pointer.
 *
 * \return The number of the audio format on success, \p -E_AUDIO_FORMAT if no
 * compiled in audio format handler is able to handler the file.
 *
 * This function tries to find an audio format handler that can interpret the
 * file given by \a data and \a size.
 *
 * It first tries to determine the audio format from the filename given by \a
 * path. If this doesn't work, all other audio format handlers are tried until
 * one is found that can handle the file.
 */
int compute_afhi(const char *path, char *data, size_t size, int fd,
		struct afh_info *afhi)
{
	int ret, i, format;

	afhi->header_len = 0;
	afhi->techinfo = NULL;
	afhi->tags.artist = NULL;
	afhi->tags.title = NULL;
	afhi->tags.year = NULL;
	afhi->tags.album = NULL;
	afhi->tags.comment = NULL;
	format = guess_audio_format(path);

	if (format >= 0) {
		ret = afl[format].get_file_info(data, size, fd, afhi);
		if (ret >= 0) {
			ret = format;
			goto success;
		}
	}
	FOR_EACH_AUDIO_FORMAT(i) {
		if (i == format) /* we already tried this one to no avail */
			continue;
		ret = afl[i].get_file_info(data, size, fd, afhi);
		if (ret >= 0) {
			ret = i;
			goto success;
		}
		PARA_WARNING_LOG("%s\n", para_strerror(-ret));
	}
	return -E_AUDIO_FORMAT;
success:
	if (!afhi->techinfo)
		afhi->techinfo = para_strdup(NULL);
	if (!afhi->tags.artist)
		afhi->tags.artist = para_strdup(NULL);
	if (!afhi->tags.title)
		afhi->tags.title = para_strdup(NULL);
	if (!afhi->tags.year)
		afhi->tags.year = para_strdup(NULL);
	if (!afhi->tags.album)
		afhi->tags.album = para_strdup(NULL);
	if (!afhi->tags.comment)
		afhi->tags.comment = para_strdup(NULL);
	PARA_DEBUG_LOG("techinfo: %s\n", afhi->techinfo);
	PARA_DEBUG_LOG("artist: %s\n", afhi->tags.artist);
	PARA_DEBUG_LOG("title: %s\n", afhi->tags.title);
	PARA_DEBUG_LOG("year: %s\n", afhi->tags.year);
	PARA_DEBUG_LOG("album: %s\n", afhi->tags.album);
	PARA_DEBUG_LOG("comment: %s\n", afhi->tags.comment);
	return ret;
}

/**
 * Get the name of the given audio format.
 *
 * \param i The audio format number.
 *
 * This returns a pointer to statically allocated memory so it
 * must not be freed by the caller.
 */
const char *audio_format_name(int i)
{
	if (i < 0 || i >= ARRAY_SIZE(afl) - 1)
		return "???";
	return afl[i].name;
}

/**
 * Get one chunk of audio data.
 *
 * \param chunk_num The number of the chunk to get.
 * \param afhi Describes the audio file.
 * \param map The memory mapped audio file.
 * \param buf Result pointer.
 * \param len The length of the chunk in bytes.
 *
 * Upon return, \a buf will point so memory inside \a map. The returned buffer
 * must therefore not be freed by the caller.
 */
void afh_get_chunk(long unsigned chunk_num, struct afh_info *afhi,
		void *map, const char **buf, size_t *len)
{
	size_t pos = afhi->chunk_table[chunk_num];
	*buf = map + pos;
	*len = afhi->chunk_table[chunk_num + 1] - pos;
}

/**
 * Get the header of an audio file.
 *
 * \param afhi The audio file handler data describing the file.
 * \param audio_format_id Determines the audio format handler.
 * \param map The data of the audio file.
 * \param mapsize The amount of bytes of the mmapped audio file.
 * \param buf The length of the header is stored here.
 * \param len Points to a buffer containing the header on return.
 *
 * This function sets \a buf to \p NULL and \a len to zero if \a map or \a
 * afhi is \p NULL, or if the current audio format does not need special
 * header treatment.
 *
 * Otherwise, it is checked whether the audio format handler given by
 * \a audio_format_id defines a ->get_header() method. If it does, this
 * method is called to obtain the header. If ->get_header() is \p NULL,
 * a reference to the first chunk of the audio file is returned.
 *
 * Once the header is no longer needed, the caller must call \ref
 * afh_free_header() to free the resources allocated by this function.
 */
void afh_get_header(struct afh_info *afhi, uint8_t audio_format_id,
		void *map, size_t mapsize, char **buf, size_t *len)
{
	struct audio_format_handler *afh = afl + audio_format_id;

	if (!map || !afhi || !afhi->header_len) {
		*buf = NULL;
		*len = 0;
		return;
	}
	if (!afh->get_header) {
		*len = afhi->header_len;
		*buf = map;
		return;
	}
	afh->get_header(map, mapsize, buf, len);
}

/**
 * Deallocate any resources obtained from afh_get_header().
 *
 * \param header_buf Pointer obtained via afh_get_header().
 * \param audio_format_id Determines the audio format handler.
 */
void afh_free_header(char *header_buf, uint8_t audio_format_id)
{
	struct audio_format_handler *afh = afl + audio_format_id;

	if (afh->get_header)
		free(header_buf);
}
