/*
 * Copyright (C) 2007-2011 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file aft.c Audio file table functions. */

#include <regex.h>
#include <dirent.h> /* readdir() */
#include <openssl/rc4.h>
#include <sys/mman.h>
#include <fnmatch.h>
#include <sys/shm.h>
#include <osl.h>

#include "para.h"
#include "error.h"
#include "crypt.h"
#include "string.h"
#include "afh.h"
#include "afs.h"
#include "net.h"
#include "fd.h"
#include "ipc.h"
#include "portable_io.h"

static struct osl_table *audio_file_table;
static char *status_items;
static char *parser_friendly_status_items;

/** The different sorting methods of the ls command. */
enum ls_sorting_method {
	/** -sp (default) */
	LS_SORT_BY_PATH,
	/** -ss */
	LS_SORT_BY_SCORE,
	/** -sl */
	LS_SORT_BY_LAST_PLAYED,
	/** -sn */
	LS_SORT_BY_NUM_PLAYED,
	/** -sf */
	LS_SORT_BY_FREQUENCY,
	/** -sc */
	LS_SORT_BY_CHANNELS,
	/** -si */
	LS_SORT_BY_IMAGE_ID,
	/** -sy */
	LS_SORT_BY_LYRICS_ID,
	/** -sb */
	LS_SORT_BY_BITRATE,
	/** -sd */
	LS_SORT_BY_DURATION,
	/** -sa */
	LS_SORT_BY_AUDIO_FORMAT,
	/** -sh */
	LS_SORT_BY_HASH,
};

/** The different listing modes of the ls command. */
enum ls_listing_mode {
	/** Default listing mode. */
	LS_MODE_SHORT,
	/** -l or -ll */
	LS_MODE_LONG,
	/** -lv */
	LS_MODE_VERBOSE,
	/** -lm */
	LS_MODE_MBOX,
	/** -lc */
	LS_MODE_CHUNKS,
	/** -lp */
	LS_MODE_PARSER,
};

/** The flags accepted by the ls command. */
enum ls_flags {
	/** -p */
	LS_FLAG_FULL_PATH = 1,
	/** -a */
	LS_FLAG_ADMISSIBLE_ONLY = 2,
	/** -r */
	LS_FLAG_REVERSE = 4,
	/** -d */
	LS_FLAG_UNIXDATE = 8,
};

/**
 * The size of the individual output fields of the ls command.
 *
 * These depend on the actual content being listed. If, for instance only files
 * with duration less than an hour are being listed, then the duration with is
 * made smaller because then the duration is listed as mm:ss rather than
 * hh:mm:ss.
 */
struct ls_widths {
	/** size of the score field. */
	unsigned short score_width;
	/** size of the image id field. */
	unsigned short image_id_width;
	/** size of the lyrics id field. */
	unsigned short lyrics_id_width;
	/** size of the bitrate field. */
	unsigned short bitrate_width;
	/** size of the frequency field. */
	unsigned short frequency_width;
	/** size of the duration field. */
	unsigned short duration_width;
	/** size of the num played field. */
	unsigned short num_played_width;
	/** size of the amp field. */
	unsigned short amp_width;
};

/** Data passed from the ls command handler to its callback function. */
struct ls_options {
	/** The given command line flags. */
	unsigned flags;
	/** The sorting method given at the command line. */
	enum ls_sorting_method sorting;
	/** The given listing mode (short, long, verbose, mbox). */
	enum ls_listing_mode mode;
	/** The arguments passed to the ls command. */
	char **patterns;
	/** Number of non-option arguments. */
	int num_patterns;
	/** Used for long listing mode to align the output fields. */
	struct ls_widths widths;
	/** Size of the \a data array. */
	uint32_t array_size;
	/** Number of used entries in the data array. */
	uint32_t num_matching_paths;
	/** Array of matching entries. */
	struct ls_data *data;
	/** Used to sort the array. */
	struct ls_data **data_ptr;
};

/**
 * Describes the layout of the mmapped-afs info struct.
 *
 * \sa struct afs_info.
 */
enum afsi_offsets {
	/** Where .last_played is stored. */
	AFSI_LAST_PLAYED_OFFSET = 0,
	/** Storage position of the attributes bitmap. */
	AFSI_ATTRIBUTES_OFFSET = 8,
	/** Storage position of the .num_played field. */
	AFSI_NUM_PLAYED_OFFSET = 16,
	/** Storage position of the .image_id field. */
	AFSI_IMAGE_ID_OFFSET = 20,
	/** Storage position of the .lyrics_id field. */
	AFSI_LYRICS_ID_OFFSET = 24,
	/** Storage position of the .audio_format_id field. */
	AFSI_AUDIO_FORMAT_ID_OFFSET = 28,
	/** Storage position of the amplification field. */
	AFSI_AMP_OFFSET = 29,
	/** 2 bytes reserved space for future usage. */
	AFSI_AUDIO_FORMAT_UNUSED_OFFSET = 30,
	/** On-disk storage space needed. */
	AFSI_SIZE = 32
};

/**
 * Convert a struct afs_info to an osl object.
 *
 * \param afsi Pointer to the audio file info to be converted.
 * \param obj Result pointer.
 *
 * \sa load_afsi().
 */
void save_afsi(struct afs_info *afsi, struct osl_object *obj)
{
	char *buf = obj->data;

	write_u64(buf + AFSI_LAST_PLAYED_OFFSET, afsi->last_played);
	write_u64(buf + AFSI_ATTRIBUTES_OFFSET, afsi->attributes);
	write_u32(buf + AFSI_NUM_PLAYED_OFFSET, afsi->num_played);
	write_u32(buf + AFSI_IMAGE_ID_OFFSET, afsi->image_id);
	write_u32(buf + AFSI_LYRICS_ID_OFFSET, afsi->lyrics_id);
	write_u8(buf + AFSI_AUDIO_FORMAT_ID_OFFSET,
		afsi->audio_format_id);
	write_u8(buf + AFSI_AMP_OFFSET, afsi->amp);
	memset(buf + AFSI_AUDIO_FORMAT_UNUSED_OFFSET, 0, 2);
}

/**
 *  Get the audio file selector info struct stored in an osl object.
 *
 * \param afsi Points to the audio_file info structure to be filled in.
 * \param obj The osl object holding the data.
 *
 * \return Positive on success, negative on errors. Possible errors: \p E_BAD_AFS.
 *
 * \sa save_afsi().
 */
int load_afsi(struct afs_info *afsi, struct osl_object *obj)
{
	char *buf = obj->data;
	if (obj->size < AFSI_SIZE)
		return -E_BAD_AFSI;
	afsi->last_played = read_u64(buf + AFSI_LAST_PLAYED_OFFSET);
	afsi->attributes = read_u64(buf + AFSI_ATTRIBUTES_OFFSET);
	afsi->num_played = read_u32(buf + AFSI_NUM_PLAYED_OFFSET);
	afsi->image_id = read_u32(buf + AFSI_IMAGE_ID_OFFSET);
	afsi->lyrics_id = read_u32(buf + AFSI_LYRICS_ID_OFFSET);
	afsi->audio_format_id = read_u8(buf +
		AFSI_AUDIO_FORMAT_ID_OFFSET);
	afsi->amp = read_u8(buf + AFSI_AMP_OFFSET);
	return 1;
}

/** The columns of the audio file table. */
enum audio_file_table_columns {
	/** The hash on the content of the audio file. */
	AFTCOL_HASH,
	/** The full path in the filesystem. */
	AFTCOL_PATH,
	/** The audio file selector info. */
	AFTCOL_AFSI,
	/** The audio format handler info. */
	AFTCOL_AFHI,
	/** The chunk table info and the chunk table of the audio file. */
	AFTCOL_CHUNKS,
	/** The number of columns of this table. */
	NUM_AFT_COLUMNS
};

/**
 * Compare two osl objects pointing to hash values.
 *
 * \param obj1 Pointer to the first hash object.
 * \param obj2 Pointer to the second hash object.
 *
 * \return The values required for an osl compare function.
 *
 * \sa osl_compare_func, uint32_compare().
 */
static int aft_hash_compare(const struct osl_object *obj1, const struct osl_object *obj2)
{
	return hash_compare((HASH_TYPE *)obj1->data, (HASH_TYPE *)obj2->data);
}

static struct osl_column_description aft_cols[] = {
	[AFTCOL_HASH] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags = OSL_RBTREE | OSL_FIXED_SIZE | OSL_UNIQUE,
		.name = "hash",
		.compare_function = aft_hash_compare,
		.data_size = HASH_SIZE
	},
	[AFTCOL_PATH] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags = OSL_RBTREE | OSL_UNIQUE,
		.name = "path",
		.compare_function = string_compare,
	},
	[AFTCOL_AFSI] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags = OSL_FIXED_SIZE,
		.name = "afs_info",
		.data_size = AFSI_SIZE
	},
	[AFTCOL_AFHI] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.name = "afh_info",
	},
	[AFTCOL_CHUNKS] = {
		.storage_type = OSL_DISK_STORAGE,
		.name = "chunks",
	}
};

static struct osl_table_description audio_file_table_desc = {
	.name = "audio_files",
	.num_columns = NUM_AFT_COLUMNS,
	.flags = OSL_LARGE_TABLE,
	.column_descriptions = aft_cols
};

/* We don't want * dot or dot-dot anywhere. */
static int verify_dotfile(const char *rest)
{
	/*
	 * The first character was '.', but that has already been discarded, we
	 * now test the rest.
	 */
	switch (*rest) {
	case '\0': case '/': /* /foo/. and /foo/./bar are not ok */
		return -1;
	case '.': /* path start with /foo/.. */
		if (rest[1] == '\0' || rest[1] == '/')
			return -1; /* /foo/.. or /foo/../bar are not ok */
		/* /foo/..bar is ok */
	}
	return 1;
}

/*
 * We fundamentally don't like some paths: We don't want double slashes or
 * slashes at the end that can make pathnames ambiguous.
 */
static int verify_path(const char *orig_path, char **resolved_path)
{
	char c;
	size_t len;
	char *path;

	if (*orig_path != '/') /* we only accept absolute paths */
		return -E_BAD_PATH;
	len = strlen(orig_path);
	*resolved_path = para_strdup(orig_path);
	path = *resolved_path;
	while (len > 1 && path[--len] == '/')
		path[len] = '\0'; /* remove slash at the end */
	c = *path++;
	while (c) {
		if (c == '/') {
			c = *path++;
			switch (c) {
			case '/': /* double slash */
				goto bad_path;
			case '.':
				if (verify_dotfile(path) < 0)
					goto bad_path;
			default:
				continue;
			}
		}
		c = *path++;
	}
	return 1;
bad_path:
	free(*resolved_path);
	return -E_BAD_PATH;
}

/** The on-disk layout of a afhi struct. */
enum afhi_offsets {
	/** Where the number of seconds is stored. */
	AFHI_SECONDS_TOTAL_OFFSET = 0,
	/** Position of the bitrate. */
	AFHI_BITRATE_OFFSET = 4,
	/** Position of the frequency. */
	AFHI_FREQUENCY_OFFSET = 8,
	/** Was: Location of the audio file header. */
	AFHI_UNUSED1_OFFSET = 12,
	/* Length of the audio file header. Zero means: No header. */
	AFHI_HEADER_LEN_OFFSET = 16,
	/** The total number of chunks (4 bytes). */
	CHUNKS_TOTAL_OFFSET = 20,
	/** The length of the audio file header (4 bytes). */
	HEADER_LEN_OFFSET = 24,
	/** Was: The start of the audio file header (4 bytes). */
	AFHI_UNUSED2_OFFSET = 28,
	/** The seconds part of the chunk time (4 bytes). */
	CHUNK_TV_TV_SEC_OFFSET = 32,
	/** The microseconds part of the chunk time (4 bytes). */
	CHUNK_TV_TV_USEC_OFFSET = 36,
	/** Number of channels is stored here. (1 byte) */
	AFHI_CHANNELS_OFFSET = 40,
	/** The tag info position. */
	AFHI_INFO_STRING_OFFSET = 41,
	/** Minimal on-disk size of a valid afhi struct. */
	MIN_AFHI_SIZE = 47, /* at least 6 null bytes for techinfo/tags */
};

static unsigned sizeof_afhi_buf(const struct afh_info *afhi)
{
	if (!afhi)
		return 0;
	return MIN_AFHI_SIZE
		+ strlen(afhi->techinfo)
		+ strlen(afhi->tags.artist)
		+ strlen(afhi->tags.title)
		+ strlen(afhi->tags.year)
		+ strlen(afhi->tags.album)
		+ strlen(afhi->tags.comment);
}

static void save_afhi(struct afh_info *afhi, char *buf)
{
	char *p;

	if (!afhi)
		return;
	write_u32(buf + AFHI_SECONDS_TOTAL_OFFSET, afhi->seconds_total);
	write_u32(buf + AFHI_BITRATE_OFFSET, afhi->bitrate);
	write_u32(buf + AFHI_FREQUENCY_OFFSET, afhi->frequency);
	write_u32(buf + AFHI_UNUSED1_OFFSET, 0);
	write_u32(buf + AFHI_HEADER_LEN_OFFSET, afhi->header_len);
	write_u8(buf + AFHI_CHANNELS_OFFSET, afhi->channels);
	write_u32(buf + CHUNKS_TOTAL_OFFSET, afhi->chunks_total);
	write_u32(buf + HEADER_LEN_OFFSET, afhi->header_len);
	write_u32(buf + AFHI_UNUSED2_OFFSET, 0);
	write_u32(buf + CHUNK_TV_TV_SEC_OFFSET, afhi->chunk_tv.tv_sec);
	write_u32(buf + CHUNK_TV_TV_USEC_OFFSET, afhi->chunk_tv.tv_usec);
	p = buf + AFHI_INFO_STRING_OFFSET;
	/* The sprintf's below are OK as our caller made sure that buf is large enough */
	p += sprintf(p, "%s", afhi->techinfo) + 1;
	p += sprintf(p, "%s", afhi->tags.artist) + 1;
	p += sprintf(p, "%s", afhi->tags.title) + 1;
	p += sprintf(p, "%s", afhi->tags.year) + 1;
	p += sprintf(p, "%s", afhi->tags.album) + 1;
	sprintf(p, "%s", afhi->tags.comment);
}

static void load_afhi(const char *buf, struct afh_info *afhi)
{
	afhi->seconds_total = read_u32(buf + AFHI_SECONDS_TOTAL_OFFSET);
	afhi->bitrate = read_u32(buf + AFHI_BITRATE_OFFSET);
	afhi->frequency = read_u32(buf + AFHI_FREQUENCY_OFFSET);
	afhi->header_len = read_u32(buf + AFHI_HEADER_LEN_OFFSET);
	afhi->channels = read_u8(buf + AFHI_CHANNELS_OFFSET);
	afhi->chunks_total = read_u32(buf + CHUNKS_TOTAL_OFFSET);
	afhi->header_len = read_u32(buf + HEADER_LEN_OFFSET);
	afhi->chunk_tv.tv_sec = read_u32(buf + CHUNK_TV_TV_SEC_OFFSET);
	afhi->chunk_tv.tv_usec = read_u32(buf + CHUNK_TV_TV_USEC_OFFSET);
	afhi->techinfo = (char *)buf + AFHI_INFO_STRING_OFFSET;
	afhi->tags.artist = afhi->techinfo + strlen(afhi->techinfo) + 1;
	afhi->tags.title = afhi->tags.artist + strlen(afhi->tags.artist) + 1;
	afhi->tags.year = afhi->tags.title + strlen(afhi->tags.title) + 1;
	afhi->tags.album = afhi->tags.year + strlen(afhi->tags.year) + 1;
	afhi->tags.comment = afhi->tags.album + strlen(afhi->tags.album) + 1;
}

static unsigned sizeof_chunk_table(struct afh_info *afhi)
{
	if (!afhi)
		return 0;
	return 4 * (afhi->chunks_total + 1);
}

static uint32_t save_chunk_table(struct afh_info *afhi, char *buf)
{
	int i;
	uint32_t max = 0, old = 0;

	for (i = 0; i <= afhi->chunks_total; i++) {
		uint32_t val = afhi->chunk_table[i];
		write_u32(buf + 4 * i, val);
		/*
		 * If the first chunk is the header, do not consider it for the
		 * calculation of the largest chunk size.
		 */
		if (i == 0 || (i == 1 && afhi->header_len > 0)) {
			old = val;
			continue;
		}
		max = PARA_MAX(max, val - old);
		old = val;
	}
	return max;
}

static void load_chunk_table(struct afh_info *afhi, char *buf)
{
	int i;

	afhi->chunk_table = para_malloc(sizeof_chunk_table(afhi));
	for (i = 0; i <= afhi->chunks_total; i++)
		afhi->chunk_table[i] = read_u32(buf + 4 * i);
}

/**
 * Get the row of the audio file table corresponding to the given path.
 *
 * \param path The full path of the audio file.
 * \param row Result pointer.
 *
 * \return Standard.
 */
int aft_get_row_of_path(const char *path, struct osl_row **row)
{
	struct osl_object obj = {.data = (char *)path, .size = strlen(path) + 1};

	return osl(osl_get_row(audio_file_table, AFTCOL_PATH, &obj, row));
}

/**
 * Get the row of the audio file table corresponding to the given hash value.
 *
 * \param hash The hash value of the desired audio file.
 * \param row Result pointer.
 *
 * \return Standard.
 */
static int aft_get_row_of_hash(HASH_TYPE *hash, struct osl_row **row)
{
	const struct osl_object obj = {.data = hash, .size = HASH_SIZE};
	return osl(osl_get_row(audio_file_table, AFTCOL_HASH, &obj, row));
}

/**
 * Get the osl object holding the audio file selector info of a row.
 *
 * \param row Pointer to a row in the audio file table.
 * \param obj Result pointer.
 *
 * \return Standard.
 */
int get_afsi_object_of_row(const struct osl_row *row, struct osl_object *obj)
{
	return osl(osl_get_object(audio_file_table, row, AFTCOL_AFSI, obj));
}

/**
 * Get the osl object holding the audio file selector info, given a path.
 *
 *
 * \param path The full path of the audio file.
 * \param obj Result pointer.
 *
 * \return Positive on success, negative on errors.
 */
static int get_afsi_object_of_path(const char *path, struct osl_object *obj)
{
	struct osl_row *row;
	int ret = aft_get_row_of_path(path, &row);
	if (ret < 0)
		return ret;
	return get_afsi_object_of_row(row, obj);
}

/**
 * Get the audio file selector info, given a row of the audio file table.
 *
 * \param row Pointer to a row in the audio file table.
 * \param afsi Result pointer.
 *
 * \return Positive on success, negative on errors.
 */
int get_afsi_of_row(const struct osl_row *row, struct afs_info *afsi)
{
	struct osl_object obj;
	int ret = get_afsi_object_of_row(row, &obj);
	if (ret < 0)
		return ret;
	return load_afsi(afsi, &obj);
}

/**
 * Get the audio file selector info, given the path of an audio table.
 *
 * \param path The full path of the audio file.
 * \param afsi Result pointer.
 *
 * \return Positive on success, negative on errors.
 */
int get_afsi_of_path(const char *path, struct afs_info *afsi)
{
	struct osl_object obj;
	int ret = get_afsi_object_of_path(path, &obj);
	if (ret < 0)
		return ret;
	return load_afsi(afsi, &obj);
}

/**
 * Get the path of an audio file, given a row of the audio file table.
 *
 * \param row Pointer to a row in the audio file table.
 * \param path Result pointer.
 *
 * The result is a pointer to mmapped data. The caller must not attempt
 * to free it.
 *
 * \return Standard.
 */
int get_audio_file_path_of_row(const struct osl_row *row, char **path)
{
	struct osl_object path_obj;
	int ret = osl(osl_get_object(audio_file_table, row, AFTCOL_PATH,
		&path_obj));
	if (ret < 0)
		return ret;
	*path = path_obj.data;
	return 1;
}

/**
 * Get the object containing the hash value of an audio file, given a row.
 *
 * \param row Pointer to a row of the audio file table.
 * \param obj Result pointer.
 *
 * \return The return value of the underlying call to osl_get_object().
 *
 * \sa get_hash_of_row().
 */
static int get_hash_object_of_aft_row(const struct osl_row *row,
		struct osl_object *obj)
{
	return osl(osl_get_object(audio_file_table, row, AFTCOL_HASH, obj));
}

/**
 * Get the hash value of an audio file, given a row of the audio file table.
 *
 * \param row Pointer to a row of the audio file table.
 * \param hash Result pointer.
 *
 * \a hash points to mapped data and must not be freed by the caller.
 *
 * \return The return value of the underlying call to
 * get_hash_object_of_aft_row().
 */
static int get_hash_of_row(const struct osl_row *row, HASH_TYPE **hash)
{
	struct osl_object obj;
	int ret = get_hash_object_of_aft_row(row, &obj);

	if (ret < 0)
		return ret;
	*hash = obj.data;
	return 1;
}

/**
 * Get the audio format handler info, given a row of the audio file table.
 *
 * \param row Pointer to a row of the audio file table.
 * \param afhi Result pointer.
 *
 * \return The return value of the underlying call to osl_get_object().
 *
 * \sa get_chunk_table_of_row().
 */
int get_afhi_of_row(const struct osl_row *row, struct afh_info *afhi)
{
	struct osl_object obj;
	int ret = osl(osl_get_object(audio_file_table, row, AFTCOL_AFHI,
		&obj));
	if (ret < 0)
		return ret;
	load_afhi(obj.data, afhi);
	return 1;
}

/* returns shmid on success */
static int save_afd(struct audio_file_data *afd)
{
	size_t size = sizeof(*afd) + sizeof_chunk_table(&afd->afhi);
	int shmid, ret = shm_new(size);
	void *shm_afd;
	char *buf;

	if (ret < 0)
		return ret;
	shmid = ret;
	ret = shm_attach(shmid, ATTACH_RW, &shm_afd);
	if (ret < 0)
		goto err;
	buf = shm_afd;
	buf += sizeof(*afd);
	afd->max_chunk_size = save_chunk_table(&afd->afhi, buf);
	*(struct audio_file_data *)shm_afd = *afd;
	shm_detach(shm_afd);
	return shmid;
err:
	shm_destroy(shmid);
	return ret;
}

/**
 * Extract a afd stored in a shared memory area.
 *
 * Attach the shared memory area given by \a shmid, load the audio file data
 * stored therein and detach the area afterwards.  Called by vss, after
 * receiving a positive response to the request for the next audio file.
 +
 * \param shmid The identifier of the shared memory area containing the afd.
 * \param afd Result pointer.
 *
 * \return Standard.
 */
int load_afd(int shmid, struct audio_file_data *afd)
{
	void *shm_afd;
	char *buf;
	int ret;

	ret = shm_attach(shmid, ATTACH_RO, &shm_afd);
	if (ret < 0)
		return ret;
	*afd = *(struct audio_file_data *)shm_afd;
	buf = shm_afd;
	buf += sizeof(*afd);
	load_chunk_table(&afd->afhi, buf);
	shm_detach(shm_afd);
	return 1;
}

static int get_local_time(uint64_t *seconds, char *buf, size_t size,
	time_t current_time, enum ls_listing_mode lm)
{
	struct tm t;

	if (!localtime_r((time_t *)seconds, &t))
		return -E_LOCALTIME;
	if (lm == LS_MODE_MBOX) {
		if (!strftime(buf, size, "%c", &t))
			return -E_STRFTIME;
		return 1;
	}
	if (*seconds + 6 * 30 * 24 * 3600 > current_time) {
		if (!strftime(buf, size, "%b %e %k:%M", &t))
			return -E_STRFTIME;
		return 1;
	}
	if (!strftime(buf, size, "%b %e  %Y", &t))
		return -E_STRFTIME;
	return 1;
}

/** Compute the number of (decimal) digits of a number. */
#define GET_NUM_DIGITS(x, num) { \
	typeof((x)) _tmp = PARA_ABS(x); \
	*num = 1; \
	if ((_tmp)) \
		while ((_tmp) > 9) { \
			(_tmp) /= 10; \
			(*num)++; \
		} \
	}

static short unsigned get_duration_width(int seconds)
{
	short unsigned width;
	unsigned hours = seconds / 3600, mins = (seconds % 3600) / 60;

	if (!hours) /* less than one hour => m:ss or mm:ss => 4 or 5 digits */
		return 4 + (mins > 9);
	/* more than one hour => h:mm:ss, hh:mm:ss, hhh:mm:ss, ... */
	GET_NUM_DIGITS(hours, &width);
	return width + 6;
}

static void get_duration_buf(int seconds, char *buf, struct ls_options *opts)
{
	unsigned hours = seconds / 3600, mins = (seconds % 3600) / 60;
	short unsigned max_width;

	if (!hours) { /* m:ss or mm:ss */
		max_width = opts->mode == LS_MODE_LONG?
			opts->widths.duration_width : 4;
		sprintf(buf, "%*u:%02u", max_width - 3, mins, seconds % 60);
	} else { /* more than one hour => h:mm:ss, hh:mm:ss, hhh:mm:ss, ... */
		max_width = opts->mode == LS_MODE_LONG?
			opts->widths.duration_width : 7;
		sprintf(buf, "%*u:%02u:%02u", max_width - 6, hours, mins,
			seconds % 60);
	}
}


static int write_attribute_items(struct para_buffer *b,
		const char *att_bitmap, struct afs_info *afsi)
{
	char *att_text;
	int ret;

	ret = WRITE_STATUS_ITEM(b, SI_ATTRIBUTES_BITMAP, "%s\n", att_bitmap);
	if (ret < 0)
		return ret;
	ret = get_attribute_text(&afsi->attributes, " ", &att_text);
	if (ret < 0)
		return ret;
	ret = WRITE_STATUS_ITEM(b, SI_ATTRIBUTES_TXT, "%s\n", att_text);
	free(att_text);
	return ret;
}

static int write_lyrics_items(struct para_buffer *b, struct afs_info *afsi)
{
	char *lyrics_name;
	int ret;

	ret = WRITE_STATUS_ITEM(b, SI_LYRICS_ID, "%u\n", afsi->lyrics_id);
	if (ret < 0)
		return ret;
	lyr_get_name_by_id(afsi->lyrics_id, &lyrics_name);
	return WRITE_STATUS_ITEM(b, SI_LYRICS_NAME, "%s\n", lyrics_name?
		lyrics_name : "(none)");
}

static int write_image_items(struct para_buffer *b, struct afs_info *afsi)
{
	char *image_name;
	int ret;

	ret = WRITE_STATUS_ITEM(b, SI_IMAGE_ID, "%u\n", afsi->image_id);
	if (ret < 0)
		return ret;
	img_get_name_by_id(afsi->image_id, &image_name);
	return WRITE_STATUS_ITEM(b, SI_IMAGE_NAME, "%s\n", image_name?
		image_name : "(none)");
}

static int write_filename_items(struct para_buffer *b, const char *path,
		unsigned flags)
{
	char *val;
	int ret;

	if (!(flags & LS_FLAG_FULL_PATH))
		return WRITE_STATUS_ITEM(b, SI_BASENAME, "%s\n", path);
	ret = WRITE_STATUS_ITEM(b, SI_PATH, "%s\n", path);
	if (ret < 0)
		return ret;
	val = para_basename(path);
	ret = WRITE_STATUS_ITEM(b, SI_BASENAME, "%s\n", val? val : "");
	if (ret < 0)
		return ret;
	val = para_dirname(path);
	ret = WRITE_STATUS_ITEM(b, SI_DIRECTORY, "%s\n", val? val : "");
	free(val);
	return ret;
}

static int print_chunk_table(struct ls_data *d, struct para_buffer *b)
{
	struct osl_object chunk_table_obj;
	struct osl_row *aft_row;
	int ret, i;
	char *buf;

	ret = aft_get_row_of_hash(d->hash, &aft_row);
	if (ret < 0)
		return ret;
	ret = osl_open_disk_object(audio_file_table, aft_row,
		AFTCOL_CHUNKS, &chunk_table_obj);
	if (ret < 0)
		return ret;
	ret = para_printf(b, "%s\n"
		"chunk_time: %lu:%lu\nchunk_offsets: ",
		d->path,
		(long unsigned) d->afhi.chunk_tv.tv_sec,
		(long unsigned) d->afhi.chunk_tv.tv_usec
	);
	if (ret < 0)
		goto out;
	buf = chunk_table_obj.data;
	for (i = 0; i <= d->afhi.chunks_total; i++) {
		ret = para_printf(b, "%u ", (unsigned) read_u32(buf + 4 * i));
		if (ret < 0)
			goto out;
	}
	ret = para_printf(b, "\n");
out:
	osl_close_disk_object(&chunk_table_obj);
	return ret;
}

static int write_score(struct para_buffer *b, struct ls_data *d,
		struct ls_options *opts)
{
	if (!(opts->flags & LS_FLAG_ADMISSIBLE_ONLY)) /* no score*/
		return 0;
	return WRITE_STATUS_ITEM(b, SI_SCORE, "%li\n", d->score);
}

static int print_list_item(struct ls_data *d, struct ls_options *opts,
	struct para_buffer *b, time_t current_time)
{
	int ret;
	char att_buf[65];
	char last_played_time[30];
	char duration_buf[30]; /* nobody has an audio file long enough to overflow this */
	struct afs_info *afsi = &d->afsi;
	struct afh_info *afhi = &d->afhi;
	char asc_hash[2 * HASH_SIZE + 1];

	if (opts->mode == LS_MODE_SHORT) {
		ret = para_printf(b, "%s\n", d->path);
		goto out;
	}
	if (opts->mode == LS_MODE_CHUNKS) {
		ret = print_chunk_table(d, b);
		goto out;
	}
	get_attribute_bitmap(&afsi->attributes, att_buf);
	if (opts->flags & LS_FLAG_UNIXDATE)
		sprintf(last_played_time, "%llu",
			(long long unsigned)afsi->last_played);
	else {
		ret = get_local_time(&afsi->last_played, last_played_time,
			sizeof(last_played_time), current_time, opts->mode);
		if (ret < 0)
			goto out;
	}
	get_duration_buf(afhi->seconds_total, duration_buf, opts);
	if (opts->mode == LS_MODE_LONG) {
		struct ls_widths *w = &opts->widths;
		if (opts->flags & LS_FLAG_ADMISSIBLE_ONLY) {
			ret = para_printf(b, "%*li ",
				opts->widths.score_width, d->score);
			if (ret < 0)
				goto out;
		}
		ret = para_printf(b,
			"%s "   /* attributes */
			"%*u "  /* amp */
			"%*d "  /* image_id  */
			"%*d "  /* lyrics_id */
			"%*d "  /* bitrate */
			"%s "   /* audio format */
			"%*d "  /* frequency */
			"%d "   /* channels */
			"%s "   /* duration */
			"%*d "  /* num_played */
			"%s "   /* last_played */
			"%s\n", /* path */
			att_buf,
			w->amp_width, afsi->amp,
			w->image_id_width, afsi->image_id,
			w->lyrics_id_width, afsi->lyrics_id,
			w->bitrate_width, afhi->bitrate,
			audio_format_name(afsi->audio_format_id),
			w->frequency_width, afhi->frequency,
			afhi->channels,
			duration_buf,
			w->num_played_width, afsi->num_played,
			last_played_time,
			d->path
		);
		goto out;
	}
	if (opts->mode == LS_MODE_MBOX) {
		const char *bn = para_basename(d->path);
		ret = para_printf(b,
			"From foo@localhost %s\n"
			"Received: from\nTo: bar\nFrom: a\n"
			"Subject: %s\n\n",
			last_played_time,
			bn? bn : "?");
		if (ret < 0)
			goto out;
	}
	ret = write_filename_items(b, d->path, opts->flags);
	if (ret < 0)
		goto out;
	ret = write_score(b, d, opts);
	if (ret < 0)
		goto out;
	ret = write_attribute_items(b, att_buf, afsi);
	if (ret < 0)
		goto out;
	ret = write_image_items(b, afsi);
	if (ret < 0)
		goto out;
	ret = write_lyrics_items(b, afsi);
	if (ret < 0)
		goto out;
	hash_to_asc(d->hash, asc_hash);
	ret = WRITE_STATUS_ITEM(b, SI_HASH, "%s\n", asc_hash);
	if (ret < 0)
		goto out;
	ret = WRITE_STATUS_ITEM(b, SI_BITRATE, "%dkbit/s\n", afhi->bitrate);
	if (ret < 0)
		goto out;
	ret = WRITE_STATUS_ITEM(b, SI_FORMAT, "%s\n",
		audio_format_name(afsi->audio_format_id));
	if (ret < 0)
		goto out;
	ret = WRITE_STATUS_ITEM(b, SI_FREQUENCY, "%dHz\n", afhi->frequency);
	if (ret < 0)
		goto out;
	ret = WRITE_STATUS_ITEM(b, SI_CHANNELS, "%d\n", afhi->channels);
	if (ret < 0)
		goto out;
	ret = WRITE_STATUS_ITEM(b, SI_DURATION, "%s\n", duration_buf);
	if (ret < 0)
		goto out;
	ret = WRITE_STATUS_ITEM(b, SI_SECONDS_TOTAL, "%lu\n",
		afhi->seconds_total);
	if (ret < 0)
		goto out;
	ret = WRITE_STATUS_ITEM(b, SI_LAST_PLAYED, "%s\n", last_played_time);
	if (ret < 0)
		goto out;
	ret = WRITE_STATUS_ITEM(b, SI_NUM_PLAYED, "%d\n", afsi->num_played);
	if (ret < 0)
		goto out;
	ret = WRITE_STATUS_ITEM(b, SI_AMPLIFICATION, "%u\n", afsi->amp);
	if (ret < 0)
		goto out;
	ret = WRITE_STATUS_ITEM(b, SI_CHUNK_TIME, "%lu\n",
		tv2ms(&afhi->chunk_tv));
	if (ret < 0)
		goto out;
	ret = WRITE_STATUS_ITEM(b, SI_NUM_CHUNKS, "%lu\n",
		afhi->chunks_total);
	if (ret < 0)
		goto out;
	ret = WRITE_STATUS_ITEM(b, SI_TECHINFO, "%s\n", afhi->techinfo);
	if (ret < 0)
		goto out;
	ret = WRITE_STATUS_ITEM(b, SI_ARTIST, "%s\n", afhi->tags.artist);
	if (ret < 0)
		goto out;
	ret = WRITE_STATUS_ITEM(b, SI_TITLE, "%s\n", afhi->tags.title);
	if (ret < 0)
		goto out;
	ret = WRITE_STATUS_ITEM(b, SI_YEAR, "%s\n", afhi->tags.year);
	if (ret < 0)
		goto out;
	ret = WRITE_STATUS_ITEM(b, SI_ALBUM, "%s\n", afhi->tags.album);
	if (ret < 0)
		goto out;
	ret = WRITE_STATUS_ITEM(b, SI_COMMENT, "%s\n", afhi->tags.comment);
	if (ret < 0)
		goto out;
	if (opts->mode == LS_MODE_MBOX) {
		struct osl_object lyrics_def;
		lyr_get_def_by_id(afsi->lyrics_id, &lyrics_def);
		if (lyrics_def.data) {
			ret = para_printf(b, "Lyrics:\n~~~~~~~\n%s",
				(char *)lyrics_def.data);
			osl_close_disk_object(&lyrics_def);
		}
	}
out:
	return ret;
}

static int make_status_items(struct audio_file_data *afd,
		struct afs_info *afsi, char *path, long score,
		HASH_TYPE *hash)
{
	struct ls_data d = {
		.afhi = afd->afhi,
		.afsi = *afsi,
		.path = path,
		.score = score,
		.hash = hash
	};
	struct ls_options opts = {
		.flags = LS_FLAG_FULL_PATH | LS_FLAG_ADMISSIBLE_ONLY,
		.mode = LS_MODE_VERBOSE,
	};
	struct para_buffer pb = {.max_size = SHMMAX - 1};
	time_t current_time;
	int ret;

	time(&current_time);
	ret = print_list_item(&d, &opts, &pb, current_time);
	if (ret < 0)
		return ret;
	free(status_items);
	status_items = pb.buf;
	memset(&pb, 0, sizeof(pb));
	pb.max_size = SHMMAX - 1;
	pb.flags = PBF_SIZE_PREFIX;
	ret = print_list_item(&d, &opts, &pb, current_time);
	if (ret < 0) {
		free(status_items);
		status_items = NULL;
		return ret;
	}
	free(parser_friendly_status_items);
	parser_friendly_status_items = pb.buf;
	return 1;
}

/**
 * Mmap the given audio file and update statistics.
 *
 * \param aft_row Determines the audio file to be opened and updated.
 * \param score The score of the audio file.
 * \param afd Result pointer.
 *
 * On success, the numplayed field of the audio file selector info is increased
 * and the lastplayed time is set to the current time. Finally, the score of
 * the audio file is updated.
 *
 * \return Positive shmid on success, negative on errors.
 */
int open_and_update_audio_file(struct osl_row *aft_row, long score,
	struct audio_file_data *afd)
{
	HASH_TYPE *aft_hash, file_hash[HASH_SIZE];
	struct osl_object afsi_obj;
	struct afs_info old_afsi, new_afsi;
	int ret = get_hash_of_row(aft_row, &aft_hash);
	struct afsi_change_event_data aced;
	struct osl_object map, chunk_table_obj;
	char *path;

	if (ret < 0)
		return ret;
	ret = get_audio_file_path_of_row(aft_row, &path);
	if (ret < 0)
		return ret;
	PARA_NOTICE_LOG("%s\n", path);
	ret = get_afsi_object_of_row(aft_row, &afsi_obj);
	if (ret < 0)
		return ret;
	ret = load_afsi(&old_afsi, &afsi_obj);
	if (ret < 0)
		return ret;
	ret = get_afhi_of_row(aft_row, &afd->afhi);
	if (ret < 0)
		return ret;
	afd->afhi.chunk_table = NULL;
	ret = osl_open_disk_object(audio_file_table, aft_row,
		AFTCOL_CHUNKS, &chunk_table_obj);
	if (ret < 0)
		goto err;
	ret = mmap_full_file(path, O_RDONLY, &map.data,
		&map.size, &afd->fd);
	if (ret < 0)
		goto err;
	hash_function(map.data, map.size, file_hash);
	ret = hash_compare(file_hash, aft_hash);
	para_munmap(map.data, map.size);
	if (ret) {
		ret = -E_HASH_MISMATCH;
		goto err;
	}
	new_afsi = old_afsi;
	new_afsi.num_played++;
	new_afsi.last_played = time(NULL);
	save_afsi(&new_afsi, &afsi_obj); /* in-place update */

	afd->audio_format_id = old_afsi.audio_format_id;
	load_chunk_table(&afd->afhi, chunk_table_obj.data);
	ret = make_status_items(afd, &old_afsi, path, score, file_hash);
	if (ret < 0)
		goto err;
	aced.aft_row = aft_row;
	aced.old_afsi = &old_afsi;
	afs_event(AFSI_CHANGE, NULL, &aced);
	ret = save_afd(afd);
err:
	free(afd->afhi.chunk_table);
	osl_close_disk_object(&chunk_table_obj);
	if (ret < 0)
		PARA_ERROR_LOG("%s: %s\n", path, para_strerror(-ret));
	return ret;
}

static int ls_audio_format_compare(const void *a, const void *b)
{
	struct ls_data *d1 = *(struct ls_data **)a, *d2 = *(struct ls_data **)b;
	return NUM_COMPARE(d1->afsi.audio_format_id, d2->afsi.audio_format_id);
}

static int ls_duration_compare(const void *a, const void *b)
{
	struct ls_data *d1 = *(struct ls_data **)a, *d2 = *(struct ls_data **)b;
	return NUM_COMPARE(d1->afhi.seconds_total, d2->afhi.seconds_total);
}

static int ls_bitrate_compare(const void *a, const void *b)
{
	struct ls_data *d1 = *(struct ls_data **)a, *d2 = *(struct ls_data **)b;
	return NUM_COMPARE(d1->afhi.bitrate, d2->afhi.bitrate);
}

static int ls_lyrics_id_compare(const void *a, const void *b)
{
	struct ls_data *d1 = *(struct ls_data **)a, *d2 = *(struct ls_data **)b;
	return NUM_COMPARE(d1->afsi.lyrics_id, d2->afsi.lyrics_id);
}

static int ls_image_id_compare(const void *a, const void *b)
{
	struct ls_data *d1 = *(struct ls_data **)a, *d2 = *(struct ls_data **)b;
	return NUM_COMPARE(d1->afsi.image_id, d2->afsi.image_id);
}

static int ls_channels_compare(const void *a, const void *b)
{
	struct ls_data *d1 = *(struct ls_data **)a, *d2 = *(struct ls_data **)b;
	return NUM_COMPARE(d1->afhi.channels, d2->afhi.channels);
}

static int ls_frequency_compare(const void *a, const void *b)
{
	struct ls_data *d1 = *(struct ls_data **)a, *d2 = *(struct ls_data **)b;
	return NUM_COMPARE(d1->afhi.frequency, d2->afhi.frequency);
}

static int ls_num_played_compare(const void *a, const void *b)
{
	struct ls_data *d1 = *(struct ls_data **)a, *d2 = *(struct ls_data **)b;
	return NUM_COMPARE(d1->afsi.num_played, d2->afsi.num_played);
}

static int ls_last_played_compare(const void *a, const void *b)
{
	struct ls_data *d1 = *(struct ls_data **)a, *d2 = *(struct ls_data **)b;
	return NUM_COMPARE(d1->afsi.last_played, d2->afsi.last_played);
}

static int ls_score_compare(const void *a, const void *b)
{
	struct ls_data *d1 = *(struct ls_data **)a, *d2 = *(struct ls_data **)b;
	return NUM_COMPARE(d1->score, d2->score);
}

static int ls_path_compare(const void *a, const void *b)
{
	struct ls_data *d1 = *(struct ls_data **)a, *d2 = *(struct ls_data **)b;
	return strcmp(d1->path, d2->path);
}

static int sort_matching_paths(struct ls_options *options)
{
	size_t nmemb = options->num_matching_paths;
	size_t size = sizeof(*options->data_ptr);
	int (*compar)(const void *, const void *);
	int i;

	options->data_ptr = para_malloc(nmemb * sizeof(*options->data_ptr));
	for (i = 0; i < nmemb; i++)
		options->data_ptr[i] = options->data + i;

	/* In these cases the array is already sorted */
	if (options->sorting == LS_SORT_BY_PATH
		&& !(options->flags & LS_FLAG_ADMISSIBLE_ONLY)
		&& (options->flags & LS_FLAG_FULL_PATH))
		return 1;
	if (options->sorting == LS_SORT_BY_SCORE &&
			options->flags & LS_FLAG_ADMISSIBLE_ONLY)
		return 1;

	switch (options->sorting) {
	case LS_SORT_BY_PATH:
		compar = ls_path_compare; break;
	case LS_SORT_BY_SCORE:
		compar = ls_score_compare; break;
	case LS_SORT_BY_LAST_PLAYED:
		compar = ls_last_played_compare; break;
	case LS_SORT_BY_NUM_PLAYED:
		compar = ls_num_played_compare; break;
	case LS_SORT_BY_FREQUENCY:
		compar = ls_frequency_compare; break;
	case LS_SORT_BY_CHANNELS:
		compar = ls_channels_compare; break;
	case LS_SORT_BY_IMAGE_ID:
		compar = ls_image_id_compare; break;
	case LS_SORT_BY_LYRICS_ID:
		compar = ls_lyrics_id_compare; break;
	case LS_SORT_BY_BITRATE:
		compar = ls_bitrate_compare; break;
	case LS_SORT_BY_DURATION:
		compar = ls_duration_compare; break;
	case LS_SORT_BY_AUDIO_FORMAT:
		compar = ls_audio_format_compare; break;
	default:
		return -E_BAD_SORT;
	}
	qsort(options->data_ptr, nmemb, size, compar);
	return 1;
}

/* row is either an aft_row or a row of the score table */
/* TODO: Only compute widths if we need them */
static int prepare_ls_row(struct osl_row *row, void *ls_opts)
{
	int ret, i;
	struct ls_options *options = ls_opts;
	struct ls_data *d;
	struct ls_widths *w;
	unsigned short num_digits;
	unsigned tmp;
	struct osl_row *aft_row;
	long score;
	char *path;

	if (options->flags & LS_FLAG_ADMISSIBLE_ONLY) {
		ret = get_score_and_aft_row(row, &score, &aft_row);
		if (ret < 0)
			return ret;
	} else
		aft_row = row;
	ret = get_audio_file_path_of_row(aft_row, &path);
	if (ret < 0)
		return ret;
	if (!(options->flags & LS_FLAG_FULL_PATH)) {
		char *p = strrchr(path, '/');
		if (p)
			path = p + 1;
	}
	if (options->num_patterns) {
		for (i = 0; i < options->num_patterns; i++) {
			ret = fnmatch(options->patterns[i], path, 0);
			if (!ret)
				break;
			if (ret == FNM_NOMATCH)
				continue;
			return -E_FNMATCH;
		}
		if (i >= options->num_patterns) /* no match */
			return 1;
	}
	tmp = options->num_matching_paths++;
	if (options->num_matching_paths > options->array_size) {
		options->array_size++;
		options->array_size *= 2;
		options->data = para_realloc(options->data, options->array_size
			* sizeof(*options->data));
	}
	d = options->data + tmp;
	ret = get_afsi_of_row(aft_row, &d->afsi);
	if (ret < 0)
		return ret;
	ret = get_afhi_of_row(aft_row, &d->afhi);
	if (ret < 0)
		return ret;
	d->path = path;
	ret = get_hash_of_row(aft_row, &d->hash);
	if (ret < 0)
		goto err;
	w = &options->widths;
	GET_NUM_DIGITS(d->afsi.image_id, &num_digits);
	w->image_id_width = PARA_MAX(w->image_id_width, num_digits);
	GET_NUM_DIGITS(d->afsi.lyrics_id, &num_digits);
	w->lyrics_id_width = PARA_MAX(w->lyrics_id_width, num_digits);
	GET_NUM_DIGITS(d->afhi.bitrate, &num_digits);
	w->bitrate_width = PARA_MAX(w->bitrate_width, num_digits);
	GET_NUM_DIGITS(d->afhi.frequency, &num_digits);
	w->frequency_width = PARA_MAX(w->frequency_width, num_digits);
	GET_NUM_DIGITS(d->afsi.num_played, &num_digits);
	w->num_played_width = PARA_MAX(w->num_played_width, num_digits);
	/* get the number of chars to print this amount of time */
	num_digits = get_duration_width(d->afhi.seconds_total);
	w->duration_width = PARA_MAX(w->duration_width, num_digits);
	GET_NUM_DIGITS(d->afsi.amp, &num_digits);
	w->amp_width = PARA_MAX(w->amp_width, num_digits);
	if (options->flags & LS_FLAG_ADMISSIBLE_ONLY) {
		GET_NUM_DIGITS(score, &num_digits);
		num_digits++; /* add one for the sign (space or "-") */
		w->score_width = PARA_MAX(w->score_width, num_digits);
		d->score = score;
	}
	return 1;
err:
	return ret;
}

static void com_ls_callback(int fd, const struct osl_object *query)
{
	struct ls_options *opts = query->data;
	char *p, *pattern_start = (char *)query->data + sizeof(*opts);
	struct para_buffer b = {.max_size = SHMMAX,
		.flags = (opts->mode == LS_MODE_PARSER)? PBF_SIZE_PREFIX : 0,
		.max_size_handler = pass_buffer_as_shm, .private_data = &fd};
	int i = 0, ret;
	time_t current_time;

	if (opts->num_patterns) {
		opts->patterns = para_malloc(opts->num_patterns * sizeof(char *));
		for (i = 0, p = pattern_start; i < opts->num_patterns; i++) {
			opts->patterns[i] = p;
			p += strlen(p) + 1;
		}
	} else
		opts->patterns = NULL;
	if (opts->flags & LS_FLAG_ADMISSIBLE_ONLY)
		ret = admissible_file_loop(opts, prepare_ls_row);
	else
		ret = osl(osl_rbtree_loop(audio_file_table, AFTCOL_PATH, opts,
			prepare_ls_row));
	if (ret < 0)
		goto out;
	if (!opts->num_matching_paths)
		goto out;
	ret = sort_matching_paths(opts);
	if (ret < 0)
		goto out;
	time(&current_time);
	if (opts->flags & LS_FLAG_REVERSE)
		for (i = opts->num_matching_paths - 1; i >= 0; i--) {
			ret = print_list_item(opts->data_ptr[i], opts, &b, current_time);
			if (ret < 0)
				goto out;
		}
	else
		for (i = 0; i < opts->num_matching_paths; i++) {
			ret = print_list_item(opts->data_ptr[i], opts, &b, current_time);
			if (ret < 0)
				goto out;
		}
out:
	if (b.offset)
		pass_buffer_as_shm(b.buf, b.offset, &fd);
	free(b.buf);
	free(opts->data);
	free(opts->data_ptr);
	free(opts->patterns);
}

/*
 * TODO: flags -h (sort by hash)
 */
int com_ls(struct rc4_context *rc4c, int argc, char * const * const argv)
{
	int i, ret;
	unsigned flags = 0;
	enum ls_sorting_method sort = LS_SORT_BY_PATH;
	enum ls_listing_mode mode = LS_MODE_SHORT;
	struct ls_options opts = {.patterns = NULL};
	struct osl_object query = {.data = &opts, .size = sizeof(opts)};

	for (i = 1; i < argc; i++) {
		const char *arg = argv[i];
		if (arg[0] != '-')
			break;
		if (!strcmp(arg, "--")) {
			i++;
			break;
		}
		if (!strncmp(arg, "-l", 2)) {
			if (!*(arg + 2)) {
				mode = LS_MODE_LONG;
				continue;
			}
			if (*(arg + 3))
				return -E_AFT_SYNTAX;
			switch(*(arg + 2)) {
			case 's':
				mode = LS_MODE_SHORT;
				continue;
			case 'l':
				mode = LS_MODE_LONG;
				continue;
			case 'v':
				mode = LS_MODE_VERBOSE;
				continue;
			case 'm':
				mode = LS_MODE_MBOX;
				continue;
			case 'c':
				mode = LS_MODE_CHUNKS;
				continue;
			case 'p':
				mode = LS_MODE_PARSER;
				continue;
			default:
				return -E_AFT_SYNTAX;
			}
		}
		if (!strcmp(arg, "-p")) {
			flags |= LS_FLAG_FULL_PATH;
			continue;
		}
		if (!strcmp(arg, "-a")) {
			flags |= LS_FLAG_ADMISSIBLE_ONLY;
			continue;
		}
		if (!strcmp(arg, "-r")) {
			flags |= LS_FLAG_REVERSE;
			continue;
		}
		if (!strcmp(arg, "-d")) {
			flags |= LS_FLAG_UNIXDATE;
			continue;
		}
		if (!strncmp(arg, "-s", 2)) {
			if (!*(arg + 2) || *(arg + 3))
				return -E_AFT_SYNTAX;
			switch(*(arg + 2)) {
			case 'p':
				sort = LS_SORT_BY_PATH;
				continue;
			case 's': /* -ss implies -a */
				sort = LS_SORT_BY_SCORE;
				flags |= LS_FLAG_ADMISSIBLE_ONLY;
				continue;
			case 'l':
				sort = LS_SORT_BY_LAST_PLAYED;
				continue;
			case 'n':
				sort = LS_SORT_BY_NUM_PLAYED;
				continue;
			case 'f':
				sort = LS_SORT_BY_FREQUENCY;
				continue;
			case 'c':
				sort = LS_SORT_BY_CHANNELS;
				continue;
			case 'i':
				sort = LS_SORT_BY_IMAGE_ID;
				continue;
			case 'y':
				sort = LS_SORT_BY_LYRICS_ID;
				continue;
			case 'b':
				sort = LS_SORT_BY_BITRATE;
				continue;
			case 'd':
				sort = LS_SORT_BY_DURATION;
				continue;
			case 'a':
				sort = LS_SORT_BY_AUDIO_FORMAT;
				continue;
			default:
				return -E_AFT_SYNTAX;
			}
		}
		return -E_AFT_SYNTAX;
	}
	opts.flags = flags;
	opts.sorting = sort;
	opts.mode = mode;
	opts.num_patterns = argc - i;
	ret = send_option_arg_callback_request(&query, opts.num_patterns,
		argv + i, com_ls_callback, rc4_send_result, rc4c);
	return ret;
}

/**
 * Call the given function for each file in the audio file table.
 *
 * \param private_data An arbitrary data pointer, passed to \a func.
 * \param func The custom function to be called.
 *
 * \return Standard.
 */
int audio_file_loop(void *private_data, osl_rbtree_loop_func *func)
{
	return osl(osl_rbtree_loop(audio_file_table, AFTCOL_HASH, private_data,
		func));
}

static struct osl_row *find_hash_sister(HASH_TYPE *hash)
{
	const struct osl_object obj = {.data = hash, .size = HASH_SIZE};
	struct osl_row *row;

	osl_get_row(audio_file_table, AFTCOL_HASH, &obj, &row);
	return row;
}

/** The format of the data stored by save_audio_file_data(). */
enum com_add_buffer_offsets {
	/** afhi (if present) starts here. */
	CAB_AFHI_OFFSET_POS = 0,
	/** Start of the chunk table (if present). */
	CAB_CHUNKS_OFFSET_POS = 2,
	/** Audio format id. */
	CAB_AUDIO_FORMAT_OFFSET = 4,
	/** Flags given to the add command. */
	CAB_FLAGS_OFFSET = 5,
	/** The hash of the audio file being added. */
	CAB_HASH_OFFSET = 9,
	/** Start of the path of the audio file. */
	CAB_PATH_OFFSET = (CAB_HASH_OFFSET + HASH_SIZE),
};

/*
 * Store the given data to a single buffer. Doesn't need the audio file selector
 * info struct as the server knows it as well.
 *
 * It's OK to call this with afhi == NULL. In this case, the audio format
 * handler info won't be stored in the buffer.
 */
static void save_add_callback_buffer(HASH_TYPE *hash, const char *path,
		struct afh_info *afhi, uint32_t flags,
		uint8_t audio_format_num, struct osl_object *obj)
{
	size_t path_len = strlen(path) + 1;
	size_t afhi_size = sizeof_afhi_buf(afhi);
	size_t size = CAB_PATH_OFFSET + path_len + afhi_size
		+ sizeof_chunk_table(afhi);
	char *buf = para_malloc(size);
	uint16_t pos;

	write_u8(buf + CAB_AUDIO_FORMAT_OFFSET, audio_format_num);
	write_u32(buf + CAB_FLAGS_OFFSET, flags);

	memcpy(buf + CAB_HASH_OFFSET, hash, HASH_SIZE);
	strcpy(buf + CAB_PATH_OFFSET, path);

	pos = CAB_PATH_OFFSET + path_len;
	PARA_DEBUG_LOG("size: %zu, afhi starts at %d\n", size, pos);
	PARA_DEBUG_LOG("last afhi byte: %p, pos %zu\n", buf + pos + afhi_size - 1,
		pos + afhi_size - 1);
	write_u16(buf + CAB_AFHI_OFFSET_POS, pos);
	save_afhi(afhi, buf + pos);

	pos += afhi_size;
	PARA_DEBUG_LOG("size: %zu, chunks start at %d\n", size, pos);
	write_u16(buf + CAB_CHUNKS_OFFSET_POS, pos);
	if (afhi)
		save_chunk_table(afhi, buf + pos);
	PARA_DEBUG_LOG("last byte in buf: %p\n", buf + size - 1);
	obj->data = buf;
	obj->size = size;
}

/*

Overview of the add command.

Input: What was passed to the callback by the command handler.
~~~~~~
HS:	Hash sister. Whether an audio file with identical hash
	already exists in the osl database.

PB:	Path brother. Whether a file with the given path exists
	in the table.

F: 	Force flag given. Whether add was called with -f.

output: Action performed by the callback.
~~~~~~~
AFHI: 	Whether afhi and chunk table are computed and sent.
ACTION:	Table modifications to be done by the callback.

+----+----+---+------+---------------------------------------------------+
| HS | PB | F | AFHI | ACTION
+----+----+---+------+---------------------------------------------------+
| Y  |  Y | Y |  Y   | if HS != PB: remove PB. HS: force afhi update,
|                    | update path, keep afsi
+----+----+---+------+---------------------------------------------------+
| Y  |  Y | N |  N   | if HS == PB: do not send callback request at all.
|                    | otherwise: remove PB, HS: update path, keep afhi,
|                    | afsi.
+----+----+---+------+---------------------------------------------------+
| Y  |  N | Y |  Y   | (rename) force afhi update of HS, update path of
|                    | HS, keep afsi
+----+----+---+------+---------------------------------------------------+
| Y  |  N | N |  N   | (file rename) update path of HS, keep afsi, afhi
+----+----+---+------+---------------------------------------------------+
| N  |  Y | Y |  Y   | (file change) update afhi, hash, of PB, keep afsi
|                    | (force has no effect)
+----+----+---+------+---------------------------------------------------+
| N  |  Y | N |  Y   | (file change) update afhi, hash of PB, keep afsi
+----+----+---+------+---------------------------------------------------+
| N  |  N | Y |  Y   | (new file) create new entry (force has no effect)
+----+----+---+------+---------------------------------------------------+
|  N |  N | N |  Y   | (new file) create new entry
+----+----+---+------+---------------------------------------------------+

Notes:

	afhi <=> force or no HS
	F => AFHI

*/

/** Flags passed to the add command. */
enum com_add_flags {
	/** Skip paths that exist already. */
	ADD_FLAG_LAZY = 1,
	/** Force adding. */
	ADD_FLAG_FORCE = 2,
	/** Print what is being done. */
	ADD_FLAG_VERBOSE = 4,
	/** Try to add files with unknown suffixes. */
	ADD_FLAG_ALL = 8,
};

static void com_add_callback(int fd, const struct osl_object *query)
{
	char *buf = query->data, *path;
	struct osl_row *pb, *aft_row;
	struct osl_row *hs;
	struct osl_object objs[NUM_AFT_COLUMNS];
	HASH_TYPE *hash;
	char asc[2 * HASH_SIZE + 1];
	int ret;
	char afsi_buf[AFSI_SIZE];
	uint32_t flags = read_u32(buf + CAB_FLAGS_OFFSET);
	struct afs_info default_afsi = {.last_played = 0};
	struct para_buffer msg = {.max_size = SHMMAX,
		.max_size_handler = pass_buffer_as_shm, .private_data = &fd};
	uint16_t afhi_offset, chunks_offset;

	hash = (HASH_TYPE *)buf + CAB_HASH_OFFSET;
	hash_to_asc(hash, asc);;
	objs[AFTCOL_HASH].data = buf + CAB_HASH_OFFSET;
	objs[AFTCOL_HASH].size = HASH_SIZE;

	path = buf + CAB_PATH_OFFSET;
	objs[AFTCOL_PATH].data = path;
	objs[AFTCOL_PATH].size = strlen(path) + 1;

	PARA_INFO_LOG("request to add %s\n", path);
	hs = find_hash_sister(hash);
	ret = aft_get_row_of_path(path, &pb);
	if (ret < 0 && ret != -OSL_ERRNO_TO_PARA_ERROR(E_OSL_RB_KEY_NOT_FOUND))
		goto out;
	if (hs && pb && hs == pb && !(flags & ADD_FLAG_FORCE)) {
		if (flags & ADD_FLAG_VERBOSE)
			ret = para_printf(&msg, "ignoring duplicate\n");
		else
			ret = 1;
		goto out;
	}
	if (hs && hs != pb) {
		struct osl_object obj;
		if (pb) { /* hs trumps pb, remove pb */
			if (flags & ADD_FLAG_VERBOSE) {
				ret = para_printf(&msg, "removing path brother\n");
				if (ret < 0)
					goto out;
			}
			ret = osl(osl_del_row(audio_file_table, pb));
			if (ret < 0)
				goto out;
			pb = NULL;
		}
		/* file rename, update hs' path */
		if (flags & ADD_FLAG_VERBOSE) {
			ret = osl(osl_get_object(audio_file_table, hs,
				AFTCOL_PATH, &obj));
			if (ret < 0)
				goto out;
			ret = para_printf(&msg, "renamed from %s\n", (char *)obj.data);
			if (ret < 0)
				goto out;
		}
		ret = osl(osl_update_object(audio_file_table, hs, AFTCOL_PATH,
			&objs[AFTCOL_PATH]));
		if (ret < 0)
			goto out;
		afs_event(AUDIO_FILE_RENAME, &msg, hs);
		if (!(flags & ADD_FLAG_FORCE))
			goto out;
	}
	/* no hs or force mode, child must have sent afhi */
	afhi_offset = read_u16(buf + CAB_AFHI_OFFSET_POS);
	chunks_offset = read_u16(buf + CAB_CHUNKS_OFFSET_POS);

	objs[AFTCOL_AFHI].data = buf + afhi_offset;
	objs[AFTCOL_AFHI].size = chunks_offset - afhi_offset;
	ret = -E_NO_AFHI;
	if (!objs[AFTCOL_AFHI].size) /* "impossible" */
		goto out;
	objs[AFTCOL_CHUNKS].data = buf + chunks_offset;
	objs[AFTCOL_CHUNKS].size = query->size - chunks_offset;
	if (pb && !hs) { /* update pb's hash */
		char old_asc[2 * HASH_SIZE + 1];
		HASH_TYPE *old_hash;
		ret = get_hash_of_row(pb, &old_hash);
		if (ret < 0)
			goto out;
		hash_to_asc(old_hash, old_asc);
		if (flags & ADD_FLAG_VERBOSE) {
			ret = para_printf(&msg, "file change: %s -> %s\n",
				old_asc, asc);
			if (ret < 0)
				goto out;
		}
		ret = osl_update_object(audio_file_table, pb, AFTCOL_HASH,
			&objs[AFTCOL_HASH]);
		if (ret < 0)
			goto out;
	}
	if (hs || pb) { /* (hs != NULL and pb != NULL) implies hs == pb */
		struct osl_row *row = pb? pb : hs;
		/* update afhi and chunk_table */
		if (flags & ADD_FLAG_VERBOSE) {
			ret = para_printf(&msg, "updating afhi and chunk table\n");
			if (ret < 0)
				goto out;
		}
		ret = osl(osl_update_object(audio_file_table, row, AFTCOL_AFHI,
			&objs[AFTCOL_AFHI]));
		if (ret < 0)
			goto out;
		ret = osl(osl_update_object(audio_file_table, row, AFTCOL_CHUNKS,
			&objs[AFTCOL_CHUNKS]));
		if (ret < 0)
			goto out;
		afs_event(AFHI_CHANGE, &msg, row);
		goto out;
	}
	/* new entry, use default afsi */
	if (flags & ADD_FLAG_VERBOSE) {
		ret = para_printf(&msg, "new file\n");
		if (ret < 0)
			goto out;
	}
	default_afsi.last_played = time(NULL) - 365 * 24 * 60 * 60;
	default_afsi.audio_format_id = read_u8(buf + CAB_AUDIO_FORMAT_OFFSET);

	objs[AFTCOL_AFSI].data = &afsi_buf;
	objs[AFTCOL_AFSI].size = AFSI_SIZE;
	save_afsi(&default_afsi, &objs[AFTCOL_AFSI]);
	ret = osl(osl_add_and_get_row(audio_file_table, objs, &aft_row));
	afs_event(AUDIO_FILE_ADD, &msg, aft_row);
out:
	if (ret < 0)
		para_printf(&msg, "%s\n", para_strerror(-ret));
	if (msg.offset)
		pass_buffer_as_shm(msg.buf, msg.offset, &fd);
	free(msg.buf);
}

/** Used by com_add(). */
struct private_add_data {
	/** The socket file descriptor, including rc4 keys. */
	struct rc4_context *rc4c;
	/** The given add flags. */
	uint32_t flags;
};

static void path_brother_callback(int fd, const struct osl_object *query)
{
	char *path = query->data;
	struct osl_row *path_brother;
	int ret = aft_get_row_of_path(path, &path_brother);
	if (ret < 0)
		return;
	pass_buffer_as_shm((char *)&path_brother, sizeof(path_brother), &fd);
}

static void hash_sister_callback(int fd, const struct osl_object *query)
{
	HASH_TYPE *hash = query->data;
	struct osl_row *hash_sister;

	hash_sister = find_hash_sister(hash);
	if (!hash_sister)
		return;
	pass_buffer_as_shm((char *)&hash_sister, sizeof(hash_sister), &fd);
}

static int get_row_pointer_from_result(struct osl_object *result, void *private)
{
	struct osl_row **row = private;
	*row = result->data;
	return 1;
}

static int add_one_audio_file(const char *path, void *private_data)
{
	int ret, send_ret = 1, fd;
	uint8_t format_num = -1;
	struct private_add_data *pad = private_data;
	struct afh_info afhi, *afhi_ptr = NULL;
	struct osl_row *pb = NULL, *hs = NULL; /* path brother/hash sister */
	struct osl_object map, obj = {.data = NULL}, query;
	HASH_TYPE hash[HASH_SIZE];

	ret = guess_audio_format(path);
	if (ret < 0 && !(pad->flags & ADD_FLAG_ALL))
		goto out_free;
	query.data = (char *)path;
	query.size = strlen(path) + 1;
	ret = send_callback_request(path_brother_callback, &query,
		get_row_pointer_from_result, &pb);
	if (ret < 0 && ret != -OSL_ERRNO_TO_PARA_ERROR(E_OSL_RB_KEY_NOT_FOUND))
		goto out_free;
	ret = 1;
	if (pb && (pad->flags & ADD_FLAG_LAZY)) { /* lazy is really cheap */
		if (pad->flags & ADD_FLAG_VERBOSE)
			send_ret = rc4_send_va_buffer(pad->rc4c,
				"lazy-ignore: %s\n", path);
		goto out_free;
	}
	/* We still want to add this file. Compute its hash. */
	ret = mmap_full_file(path, O_RDONLY, &map.data, &map.size, &fd);
	if (ret < 0)
		goto out_free;
	hash_function(map.data, map.size, hash);

	/* Check whether the database contains a file with the same hash. */
	query.data = hash;
	query.size = HASH_SIZE;
	ret = send_callback_request(hash_sister_callback, &query,
		get_row_pointer_from_result, &hs);
	if (ret < 0)
		goto out_unmap;
	/* Return success if we already know this file. */
	ret = 1;
	if (pb && hs && hs == pb && !(pad->flags & ADD_FLAG_FORCE)) {
		if (pad->flags & ADD_FLAG_VERBOSE)
			send_ret = rc4_send_va_buffer(pad->rc4c,
				"%s exists, not forcing update\n", path);
		goto out_unmap;
	}
	/*
	 * We won't recalculate the audio format info and the chunk table if
	 * there is a hash sister and FORCE was not given.
	 */
	if (!hs || (pad->flags & ADD_FLAG_FORCE)) {
		ret = compute_afhi(path, map.data, map.size, fd, &afhi);
		if (ret < 0)
			goto out_unmap;
		format_num = ret;
		afhi_ptr = &afhi;
	}
	munmap(map.data, map.size);
	close(fd);
	if (pad->flags & ADD_FLAG_VERBOSE) {
		send_ret = rc4_send_va_buffer(pad->rc4c, "adding %s\n", path);
		if (send_ret < 0)
			goto out_free;
	}
	save_add_callback_buffer(hash, path, afhi_ptr, pad->flags, format_num, &obj);
	/* Ask afs to consider this entry for adding. */
	ret = send_callback_request(com_add_callback, &obj, rc4_send_result, pad->rc4c);
	goto out_free;

out_unmap:
	close(fd);
	munmap(map.data, map.size);
out_free:
	if (ret < 0 && send_ret >= 0)
		send_ret = rc4_send_va_buffer(pad->rc4c,
			"failed to add %s (%s)\n", path, para_strerror(-ret));
	free(obj.data);
	if (afhi_ptr) {
		free(afhi_ptr->chunk_table);
		free(afhi_ptr->techinfo);
		free(afhi_ptr->tags.artist);
		free(afhi_ptr->tags.title);
		free(afhi_ptr->tags.year);
		free(afhi_ptr->tags.album);
		free(afhi_ptr->tags.comment);
	}
	/* Stop adding files only on send errors. */
	return send_ret;
}

int com_add(struct rc4_context *rc4c, int argc, char * const * const argv)
{
	int i, ret;
	struct private_add_data pad = {.rc4c = rc4c, .flags = 0};
	struct stat statbuf;

	for (i = 1; i < argc; i++) {
		const char *arg = argv[i];
		if (arg[0] != '-')
			break;
		if (!strcmp(arg, "--")) {
			i++;
			break;
		}
		if (!strcmp(arg, "-a")) {
			pad.flags |= ADD_FLAG_ALL;
			continue;
		}
		if (!strcmp(arg, "-l")) {
			pad.flags |= ADD_FLAG_LAZY;
			continue;
		}
		if (!strcmp(arg, "-f")) {
			pad.flags |= ADD_FLAG_FORCE;
			continue;
		}
		if (!strcmp(arg, "-v")) {
			pad.flags |= ADD_FLAG_VERBOSE;
			continue;
		}
	}
	if (argc <= i)
		return -E_AFT_SYNTAX;
	for (; i < argc; i++) {
		char *path;
		ret = verify_path(argv[i], &path);
		if (ret < 0) {
			ret = rc4_send_va_buffer(rc4c, "%s: %s\n", argv[i],
				para_strerror(-ret));
			if (ret < 0)
				return ret;
			continue;
		}
		ret = stat(path, &statbuf);
		if (ret < 0) {
			ret = rc4_send_va_buffer(rc4c, "failed to stat %s (%s)\n", path,
				strerror(errno));
			free(path);
			if (ret < 0)
				return ret;
			continue;
		}
		if (S_ISDIR(statbuf.st_mode))
			ret = for_each_file_in_dir(path, add_one_audio_file,
				&pad);
		else
			ret = add_one_audio_file(path, &pad);
		if (ret < 0) {
			rc4_send_va_buffer(rc4c, "%s: %s\n", path, para_strerror(-ret));
			free(path);
			return ret;
		}
		free(path);
	}
	return 1;

}

/**
 * Flags used by the touch command.
 *
 * \sa com_touch().
 */
enum touch_flags {
	/** Whether the \p FNM_PATHNAME flag should be passed to fnmatch(). */
	TOUCH_FLAG_FNM_PATHNAME = 1,
	/** Activates verbose mode. */
	TOUCH_FLAG_VERBOSE = 2
};

/** Options used by com_touch(). */
struct com_touch_options {
	/** New num_played value. */
	int32_t num_played;
	/** New last played count. */
	int64_t last_played;
	/** New lyrics id. */
	int32_t lyrics_id;
	/** New image id. */
	int32_t image_id;
	/** New amplification value. */
	int32_t amp;
	/** Command line flags (see \ref touch_flags). */
	unsigned flags;
};

/** Data passed to the action handler of com_touch(). */
struct touch_action_data {
	/** Command line options (see \ref com_touch_options). */
	struct com_touch_options *cto;
	/** Message buffer. */
	struct para_buffer pb;
	/** How many audio files matched the given pattern. */
	unsigned num_matches;
};

static int touch_audio_file(__a_unused struct osl_table *table,
		struct osl_row *row, const char *name, void *data)
{
	struct touch_action_data *tad = data;
	struct osl_object obj;
	struct afs_info old_afsi, new_afsi;
	int ret, no_options = tad->cto->num_played < 0 && tad->cto->last_played < 0 &&
		tad->cto->lyrics_id < 0 && tad->cto->image_id < 0 && tad->cto->amp < 0;
	struct afsi_change_event_data aced;

	ret = get_afsi_object_of_row(row, &obj);
	if (ret < 0)
		return para_printf(&tad->pb, "%s: %s\n", name, para_strerror(-ret));
	ret = load_afsi(&old_afsi, &obj);
	if (ret < 0)
		return para_printf(&tad->pb, "%s: %s\n", name, para_strerror(-ret));
	new_afsi = old_afsi;
	if (no_options) {
		new_afsi.num_played++;
		new_afsi.last_played = time(NULL);
		if (tad->cto->flags & TOUCH_FLAG_VERBOSE) {
			ret = para_printf(&tad->pb, "%s: num_played = %u, "
				"last_played = now()\n", name,
				new_afsi.num_played);
			if (ret < 0)
				return ret;
		}
	} else {
		if (tad->cto->flags & TOUCH_FLAG_VERBOSE) {
			ret = para_printf(&tad->pb, "touching %s\n", name);
			if (ret < 0)
				return ret;
		}
		if (tad->cto->lyrics_id >= 0)
			new_afsi.lyrics_id = tad->cto->lyrics_id;
		if (tad->cto->image_id >= 0)
			new_afsi.image_id = tad->cto->image_id;
		if (tad->cto->num_played >= 0)
			new_afsi.num_played = tad->cto->num_played;
		if (tad->cto->last_played >= 0)
			new_afsi.last_played = tad->cto->last_played;
		if (tad->cto->amp >= 0)
			new_afsi.amp = tad->cto->amp;
	}
	tad->num_matches++;
	save_afsi(&new_afsi, &obj); /* in-place update */
	aced.aft_row = row;
	aced.old_afsi = &old_afsi;
	afs_event(AFSI_CHANGE, &tad->pb, &aced);
	return 1;
}

static void com_touch_callback(int fd, const struct osl_object *query)
{
	struct touch_action_data tad = {.cto = query->data,
		.pb = {
			.max_size = SHMMAX,
			.private_data = &fd,
			.max_size_handler = pass_buffer_as_shm
		}
	};
	int ret, ret2 = 0;
	struct pattern_match_data pmd = {
		.table = audio_file_table,
		.loop_col_num = AFTCOL_HASH,
		.match_col_num = AFTCOL_PATH,
		.patterns = {.data = (char *)query->data + sizeof(*tad.cto),
			.size = query->size - sizeof(*tad.cto)},
		.data = &tad,
		.action = touch_audio_file
	};
	if (tad.cto->flags & TOUCH_FLAG_FNM_PATHNAME)
		pmd.fnmatch_flags |= FNM_PATHNAME;
	ret = for_each_matching_row(&pmd);
	if (ret < 0)
		ret2 = para_printf(&tad.pb, "%s\n", para_strerror(-ret));
	else
		if (!tad.num_matches)
			ret2 = para_printf(&tad.pb, "no matches\n");
	if (ret2 >= 0 && tad.pb.offset)
		pass_buffer_as_shm(tad.pb.buf, tad.pb.offset, &fd);
	free(tad.pb.buf);
}

int com_touch(struct rc4_context *rc4c, int argc, char * const * const argv)
{
	struct com_touch_options cto = {
		.num_played = -1,
		.last_played = -1,
		.lyrics_id = -1,
		.image_id = -1,
		.amp = -1,
	};
	struct osl_object query = {.data = &cto, .size = sizeof(cto)};
	int i, ret;


	for (i = 1; i < argc; i++) {
		const char *arg = argv[i];
		if (arg[0] != '-')
			break;
		if (!strcmp(arg, "--")) {
			i++;
			break;
		}
		if (!strncmp(arg, "-n=", 3)) {
			ret = para_atoi32(arg + 3, &cto.num_played);
			if (ret < 0)
				return ret;
			continue;
		}
		if (!strncmp(arg, "-l=", 3)) {
			ret = para_atoi64(arg + 3, &cto.last_played);
			if (ret < 0)
				return ret;
			continue;
		}
		if (!strncmp(arg, "-y=", 3)) {
			ret = para_atoi32(arg + 3, &cto.lyrics_id);
			if (ret < 0)
				return ret;
			continue;
		}
		if (!strncmp(arg, "-i=", 3)) {
			ret = para_atoi32(arg + 3, &cto.image_id);
			if (ret < 0)
				return ret;
			continue;
		}
		if (!strncmp(arg, "-a=", 3)) {
			int32_t val;
			ret = para_atoi32(arg + 3, &val);
			if (ret < 0)
				return ret;
			if (val < 0 || val > 255)
				return -ERRNO_TO_PARA_ERROR(EINVAL);
			cto.amp = val;
			continue;
		}
		if (!strcmp(arg, "-p")) {
			cto.flags |= TOUCH_FLAG_FNM_PATHNAME;
			continue;
		}
		if (!strcmp(arg, "-v")) {
			cto.flags |= TOUCH_FLAG_VERBOSE;
			continue;
		}
		break; /* non-option starting with dash */
	}
	if (i >= argc)
		return -E_AFT_SYNTAX;
	ret = send_option_arg_callback_request(&query, argc - i,
		argv + i, com_touch_callback, rc4_send_result, rc4c);
	if (ret < 0)
		rc4_send_va_buffer(rc4c, "%s\n", para_strerror(-ret));
	return ret;
}

/** Flags for com_rm(). */
enum rm_flags {
	/** -v */
	RM_FLAG_VERBOSE = 1,
	/** -f */
	RM_FLAG_FORCE = 2,
	/** -p */
	RM_FLAG_FNM_PATHNAME = 4
};

/** Data passed to the action handler of com_rm(). */
struct com_rm_action_data {
	/** Command line flags ((see \ref rm_flags). */
	uint32_t flags;
	/** Message buffer. */
	struct para_buffer pb;
	/** Number of audio files removed. */
	unsigned num_removed;
};

static int remove_audio_file(__a_unused struct osl_table *table,
		struct osl_row *row, const char *name, void *data)
{
	struct com_rm_action_data *crd = data;
	int ret;

	if (crd->flags & RM_FLAG_VERBOSE) {
		ret = para_printf(&crd->pb, "removing %s\n", name);
		if (ret < 0)
			return ret;
	}
	afs_event(AUDIO_FILE_REMOVE, &crd->pb, row);
	ret = osl(osl_del_row(audio_file_table, row));
	if (ret < 0)
		para_printf(&crd->pb, "%s: %s\n", name, para_strerror(-ret));
	else
		crd->num_removed++;
	return ret;
}

static void com_rm_callback(int fd, const struct osl_object *query)
{
	struct com_rm_action_data crd = {.flags = *(uint32_t *)query->data,
		.pb = {
			.max_size = SHMMAX,
			.private_data = &fd,
			.max_size_handler = pass_buffer_as_shm
		}
	};
	int ret;
	struct pattern_match_data pmd = {
		.table = audio_file_table,
		.loop_col_num = AFTCOL_HASH,
		.match_col_num = AFTCOL_PATH,
		.patterns = {.data = (char *)query->data + sizeof(uint32_t),
			.size = query->size - sizeof(uint32_t)},
		.data = &crd,
		.action = remove_audio_file
	};
	if (crd.flags & RM_FLAG_FNM_PATHNAME)
		pmd.fnmatch_flags |= FNM_PATHNAME;
	ret = for_each_matching_row(&pmd);
	if (ret < 0) {
		para_printf(&crd.pb, "%s\n", para_strerror(-ret));
		return;
	}
	if (!crd.num_removed && !(crd.flags & RM_FLAG_FORCE))
		ret = para_printf(&crd.pb, "no matches -- nothing removed\n");
	else {
		if (crd.flags & RM_FLAG_VERBOSE)
			ret = para_printf(&crd.pb, "removed %u files\n", crd.num_removed);
	}
	if (ret >= 0 && crd.pb.offset)
		pass_buffer_as_shm(crd.pb.buf, crd.pb.offset, &fd);
	free(crd.pb.buf);
}

/* TODO options: -r (recursive) */
int com_rm(struct rc4_context *rc4c, int argc,  char * const * const argv)
{
	uint32_t flags = 0;
	struct osl_object query = {.data = &flags, .size = sizeof(flags)};
	int i, ret;

	for (i = 1; i < argc; i++) {
		const char *arg = argv[i];
		if (arg[0] != '-')
			break;
		if (!strcmp(arg, "--")) {
			i++;
			break;
		}
		if (!strcmp(arg, "-f")) {
			flags |= RM_FLAG_FORCE;
			continue;
		}
		if (!strcmp(arg, "-p")) {
			flags |= RM_FLAG_FNM_PATHNAME;
			continue;
		}
		if (!strcmp(arg, "-v")) {
			flags |= RM_FLAG_VERBOSE;
			continue;
		}
		break;
	}
	if (i >= argc)
		return -E_AFT_SYNTAX;
	ret = send_option_arg_callback_request(&query, argc - i, argv + i,
		com_rm_callback, rc4_send_result, rc4c);
	if (ret < 0)
		rc4_send_va_buffer(rc4c, "%s\n", para_strerror(-ret));
	return ret;
}

/**
 * Flags used by the cpsi command.
 *
 * \sa com_cpsi().
 */
enum cpsi_flags {
	/** Whether the lyrics id should be copied. */
	CPSI_FLAG_COPY_LYRICS_ID = 1,
	/** Whether the image id should be copied. */
	CPSI_FLAG_COPY_IMAGE_ID = 2,
	/** Whether the lastplayed time should be copied. */
	CPSI_FLAG_COPY_LASTPLAYED = 4,
	/** Whether the numplayed count should be copied. */
	CPSI_FLAG_COPY_NUMPLAYED = 8,
	/** Whether the attributes should be copied. */
	CPSI_FLAG_COPY_ATTRIBUTES = 16,
	/** Activates verbose mode. */
	CPSI_FLAG_VERBOSE = 32,
};

/** Data passed to the action handler of com_cpsi(). */
struct cpsi_action_data {
	/** command line flags (see \ref cpsi_flags). */
	unsigned flags;
	/** Number of audio files changed. */
	unsigned num_copied;
	/** Message buffer. */
	struct para_buffer pb;
	/** Values are copied from here. */
	struct afs_info source_afsi;
};

static int copy_selector_info(__a_unused struct osl_table *table,
		struct osl_row *row, const char *name, void *data)
{
	struct cpsi_action_data *cad = data;
	struct osl_object target_afsi_obj;
	int ret;
	struct afs_info old_afsi, target_afsi;
	struct afsi_change_event_data aced;

	ret = get_afsi_object_of_row(row, &target_afsi_obj);
	if (ret < 0)
		return ret;
	load_afsi(&target_afsi, &target_afsi_obj);
	old_afsi = target_afsi;
	if (cad->flags & CPSI_FLAG_COPY_LYRICS_ID)
		target_afsi.lyrics_id = cad->source_afsi.lyrics_id;
	if (cad->flags & CPSI_FLAG_COPY_IMAGE_ID)
		target_afsi.image_id = cad->source_afsi.image_id;
	if (cad->flags & CPSI_FLAG_COPY_LASTPLAYED)
		target_afsi.last_played = cad->source_afsi.last_played;
	if (cad->flags & CPSI_FLAG_COPY_NUMPLAYED)
		target_afsi.num_played = cad->source_afsi.num_played;
	if (cad->flags & CPSI_FLAG_COPY_ATTRIBUTES)
		target_afsi.attributes = cad->source_afsi.attributes;
	save_afsi(&target_afsi, &target_afsi_obj); /* in-place update */
	cad->num_copied++;
	if (cad->flags & CPSI_FLAG_VERBOSE) {
		ret = para_printf(&cad->pb, "copied afsi to %s\n", name);
		if (ret < 0)
			return ret;
	}
	aced.aft_row = row;
	aced.old_afsi = &old_afsi;
	afs_event(AFSI_CHANGE, &cad->pb, &aced);
	return 1;
}

static void com_cpsi_callback(int fd, const struct osl_object *query)
{
	struct cpsi_action_data cad = {
		.flags = *(unsigned *)query->data,
		.pb = {
			.max_size = SHMMAX,
			.private_data = &fd,
			.max_size_handler = pass_buffer_as_shm
		}
	};
	int ret;
	char *source_path = (char *)query->data + sizeof(cad.flags);
	struct pattern_match_data pmd = {
		.table = audio_file_table,
		.loop_col_num = AFTCOL_HASH,
		.match_col_num = AFTCOL_PATH,
		.patterns = {.data = source_path + strlen(source_path) + 1,
			.size = query->size - sizeof(cad.flags)
				- strlen(source_path) - 1},
		.data = &cad,
		.action = copy_selector_info
	};

	ret = get_afsi_of_path(source_path, &cad.source_afsi);
	if (ret < 0)
		goto out;
	ret = for_each_matching_row(&pmd);
out:
	if (ret < 0)
		para_printf(&cad.pb, "%s\n", para_strerror(-ret));
	else if (cad.flags & CPSI_FLAG_VERBOSE) {
		if (cad.num_copied)
			para_printf(&cad.pb, "copied requested afsi from %s "
				"to %u files\n", source_path, cad.num_copied);
		else
			para_printf(&cad.pb, "nothing copied\n");
	}
	if (cad.pb.offset)
		pass_buffer_as_shm(cad.pb.buf, cad.pb.offset, &fd);
	free(cad.pb.buf);
}

int com_cpsi(struct rc4_context *rc4c, int argc,  char * const * const argv)
{
	unsigned flags = 0;
	int i, ret;
	struct osl_object options = {.data = &flags, .size = sizeof(flags)};

	for (i = 1; i < argc; i++) {
		const char *arg = argv[i];
		if (arg[0] != '-')
			break;
		if (!strcmp(arg, "--")) {
			i++;
			break;
		}
		if (!strcmp(arg, "-y")) {
			flags |= CPSI_FLAG_COPY_LYRICS_ID;
			continue;
		}
		if (!strcmp(arg, "-i")) {
			flags |= CPSI_FLAG_COPY_IMAGE_ID;
			continue;
		}
		if (!strcmp(arg, "-l")) {
			flags |= CPSI_FLAG_COPY_LASTPLAYED;
			continue;
		}
		if (!strcmp(arg, "-n")) {
			flags |= CPSI_FLAG_COPY_NUMPLAYED;
			continue;
		}
		if (!strcmp(arg, "-a")) {
			flags |= CPSI_FLAG_COPY_ATTRIBUTES;
			continue;
		}
		if (!strcmp(arg, "-v")) {
			flags |= CPSI_FLAG_VERBOSE;
			continue;
		}
		break;
	}
	if (i + 1 >= argc) /* need at least source file and pattern */
		return -E_AFT_SYNTAX;
	if (!(flags & ~CPSI_FLAG_VERBOSE)) /* no copy flags given */
		flags = ~(unsigned)CPSI_FLAG_VERBOSE | flags;
	ret = send_option_arg_callback_request(&options, argc - i, argv + i,
		com_cpsi_callback, rc4_send_result, rc4c);
	if (ret < 0)
		rc4_send_va_buffer(rc4c, "%s\n", para_strerror(-ret));
	return ret;
}

static void afs_stat_callback(int fd, const struct osl_object *query)
{
	int *parser_friendly = query->data;
	char *buf = *parser_friendly?
		parser_friendly_status_items : status_items;

	if (!buf)
		return;
	pass_buffer_as_shm(buf, strlen(buf), &fd);
}

/**
 * Get the current afs status items from the afs process and send it using RC4.
 *
 * \param rc4c The rc4 context for data encryption.
 * \param parser_friendly Whether parser-friendly output format should be used.
 *
 * As the contents of the afs status items change in time and the command
 * handler only has a COW version created at fork time, it can not send
 * up-to-date afs status items directly. Therefore the usual callback mechanism
 * is used to pass the status items from the afs process to the command handler
 * via a shared memory area and a pipe.
 */
int send_afs_status(struct rc4_context *rc4c, int parser_friendly)
{
	struct osl_object query = {.data = &parser_friendly,
		.size = sizeof(parser_friendly)};

	return send_callback_request(afs_stat_callback, &query, rc4_send_result, rc4c);
}

/* TODO: optionally fix problems by removing offending rows */
static int check_audio_file(struct osl_row *row, void *data)
{
	char *path;
	struct para_buffer *pb = data;
	struct stat statbuf;
	int ret = get_audio_file_path_of_row(row, &path);
	struct afs_info afsi;
	char *blob_name;

	if (ret < 0)
		return para_printf(pb, "%s\n", para_strerror(-ret));
	if (stat(path, &statbuf) < 0) {
		ret = para_printf(pb, "%s: stat error (%s)\n", path, strerror(errno));
		if (ret < 0)
			return ret;
	} else {
		if (!S_ISREG(statbuf.st_mode)) {
			ret = para_printf(pb, "%s: not a regular file\n", path);
			if (ret < 0)
				return ret;
		}
	}
	ret = get_afsi_of_row(row, &afsi);
	if (ret < 0)
		return para_printf(pb, "%s: %s\n", path, para_strerror(-ret));
	ret = lyr_get_name_by_id(afsi.lyrics_id, &blob_name);
	if (ret < 0) {
		ret = para_printf(pb, "%s lyrics id %u: %s\n", path, afsi.lyrics_id,
			para_strerror(-ret));
		if (ret < 0)
			return ret;
	}
	ret = img_get_name_by_id(afsi.image_id, &blob_name);
	if (ret < 0)
		ret = para_printf(pb, "%s image id %u: %s\n", path, afsi.image_id,
			para_strerror(-ret));
	return ret;
}

/**
 * Check the audio file table for inconsistencies.
 *
 * \param fd The afs socket.
 * \param query Unused.
 *
 * This function always succeeds.
 *
 * \sa com_check().
 */
void aft_check_callback(int fd, __a_unused const struct osl_object *query)
{
	struct para_buffer pb = {
		.max_size = SHMMAX,
		.private_data = &fd,
		.max_size_handler = pass_buffer_as_shm
	};
	int ret = para_printf(&pb, "checking audio file table...\n");

	if (ret < 0)
		return;
	audio_file_loop(&pb, check_audio_file);
	if (pb.offset)
		pass_buffer_as_shm(pb.buf, pb.offset, &fd);
	free(pb.buf);
}

/**
 * Close the audio file table.
 *
 * \param flags Usual flags that are passed to osl_close_table().
 *
 * \sa osl_close_table().
 */
static void aft_close(void)
{
	osl_close_table(audio_file_table, OSL_MARK_CLEAN);
	audio_file_table = NULL;
	free(status_items);
	status_items = NULL;
	free(parser_friendly_status_items);
	parser_friendly_status_items = NULL;
}

/**
 * Open the audio file table.
 *
 * \param dir The database directory.
 *
 * \return Standard.
 *
 * \sa osl_open_table().
 */
static int aft_open(const char *dir)
{
	int ret;

	audio_file_table_desc.dir = dir;
	ret = osl(osl_open_table(&audio_file_table_desc, &audio_file_table));
	if (ret >= 0) {
		unsigned num;
		osl_get_num_rows(audio_file_table, &num);
		PARA_INFO_LOG("audio file table contains %d files\n", num);
		return ret;
	}
	PARA_INFO_LOG("failed to open audio file table\n");
	audio_file_table = NULL;
	if (ret >= 0 || ret == -OSL_ERRNO_TO_PARA_ERROR(E_OSL_NOENT))
		return 1;
	return ret;
}

static int aft_create(const char *dir)
{
	audio_file_table_desc.dir = dir;
	return osl(osl_create_table(&audio_file_table_desc));
}

static int clear_attribute(struct osl_row *row, void *data)
{
	struct rmatt_event_data *red = data;
	struct afs_info afsi;
	struct osl_object obj;
	int ret = get_afsi_object_of_row(row, &obj);
	uint64_t mask = ~(1ULL << red->bitnum);

	if (ret < 0)
		return ret;
	ret = load_afsi(&afsi, &obj);
	if (ret < 0)
		return ret;
	afsi.attributes &= mask;
	save_afsi(&afsi, &obj);
	return 1;
}

static int aft_event_handler(enum afs_events event, struct para_buffer *pb,
		void *data)
{
	int ret;

	switch(event) {
	case ATTRIBUTE_REMOVE: {
		const struct rmatt_event_data *red = data;
		ret = para_printf(pb, "clearing attribute %s (bit %u) from all "
			"entries in the audio file table\n", red->name,
			red->bitnum);
		if (ret < 0)
			return ret;
		return audio_file_loop(data, clear_attribute);
		}
	default:
		return 1;
	}
}

/**
 * Initialize the audio file table.
 *
 * \param t Pointer to the structure to be initialized.
 */
void aft_init(struct afs_table *t)
{
	t->open = aft_open;
	t->close = aft_close;
	t->create = aft_create;
	t->event_handler = aft_event_handler;
}
