/*
 * Copyright (C) 2005-2011 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file filter.h Filter-related structures and exported symbols from filter_common.c. */

/** The list of supported filters. */
enum filter_enum {FILTER_ENUM};

/**
 * Describes one running instance of a filter.
*/
struct filter_node {
	/** The number in the array of available filters. */
	unsigned filter_num;
	/**
	 * Each filter may store any filter-specific information about the particular
	 * instance of the filter here.
	 */
	void *private_data;
	/** The list of registered callbacks. */
	struct list_head callbacks;
	/** A pointer to the configuration of this instance. */
	void *conf;
	struct btr_node *btrn;
	struct task task;
	size_t min_iqs;
};

/**
 * The structure associated with a paraslash filter.
 *
 * Paraslash filters are "modules" which are used to transform an audio stream.
 * struct filter contains pointers to functions that must be supplied by the
 * filter code in order to be used by the driving application (currently
 * para_audiod and para_filter).
 *
 * Note: As several instances of the same filter may be running at the same
 * time, all these filter functions must be reentrant; no static non-constant
 * variables may be used.
 * \sa mp3dec_filter.c, oggdec_filter.c, wav_filter.c, compress_filter.c, filter_node
 */
struct filter {
	/** The name of the filter. */
	const char *name;
	/**
	 * Pointer to the filter init routine.
	 *
	 * This function is only called once at startup. It must initialize the
	 * other non-optional function pointers of this structure.
	 */
	void (*init)(struct filter *f);
	/**
	 * Open one instance of this filter.
	 *
	 * This should allocate the output buffer of the given filter node and do any
	 * other filter-specific preparations like initializing the private_data member
	 * of \a fn suitably. The open function is assumed to succeed.
	 */
	void (*open)(struct filter_node *fn);
	/**
	 * Close one instance of this filter.
	 *
	 * Free all resources of associated with \a fn that were previously allocated
	 * by the open() function. It's OK to leave this alone if the filter does not
	 * need any cleanups.
	 */
	void (*close)(struct filter_node *fn);
	/**
	 * A pointer to the filter's command line parser.
	 *
	 * If this optional function pointer is not NULL, any filter options
	 * are passed from the main program to this command line parser once at
	 * application startup. The command line parser should check its
	 * command line options given by \a argc and \a argv and abort on
	 * errors. Success must be indicated by a non-negative return value. In
	 * this case the function should return a pointer to the
	 * filter-specific configuration data determined by \a argc and \a
	 * argv. On failure, a negative paraslash error code must be returned.
	 */
	int (*parse_config)(int argc, char **argv, void **config);
	/**
	 * Deallocate the memory for the configuration.
	 *
	 * This is called to free whatever ->parse_config() has allocated.
	 */
	void (*free_config)(void *conf);

	/** The help texts for this filter. */
	struct ggo_help help;
	/**
	 * Set scheduler timeout and add file descriptors to fd sets.
	 *
	 * This function is used to control the timeout value for select. It
	 * only allowed to decrease the current value. The second purpose of
	 * this function is to set file descriptors to be watched by the
	 * subsequent select call to the two fd sets.
	 */
	void (*pre_select)(struct sched *s, struct task *t);
	/**
	 * Convert (filter) the given data.
	 *
	 * Pointer to the converting function of the filter. On errors, the
	 * post_select function is supposed to set t->error to a (negative)
	 * error code.
	 */
	void (*post_select)(struct sched *s, struct task *t);
	/**
	 * Answer a buffer tree query.
	 *
	 * This optional function pointer is used for inter node communications
	 * of the buffer tree nodes. See \ref btr_command_handler for details.
	 */
	btr_command_handler execute;
};

void filter_init(void);
int check_filter_arg(char *filter_arg, void **conf);
void print_filter_helps(int detailed);
void generic_filter_pre_select(struct sched *s, struct task *t);
int decoder_execute(const char *cmd, unsigned sample_rate, unsigned channels,
		char **result);

static inline void write_int16_host_endian(char *buf, int val)
{
#ifdef WORDS_BIGENDIAN
	*buf = val >> 8;
	*(buf + 1) = val & 0xff;
#else
	*buf = val & 0xff;
	*(buf + 1) = val >> 8;
#endif
}

DECLARE_FILTER_INITS

#define FOR_EACH_SUPPORTED_FILTER(j)  for (j = 0; j < NUM_SUPPORTED_FILTERS; j++)

/** The filter array, one structure for each supported filter. */
extern struct filter filters[NUM_SUPPORTED_FILTERS];
