/*
 * Copyright (C) 2007-2011 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file mood.c Paraslash's mood handling functions. */

#include <regex.h>
#include <osl.h>

#include "para.h"
#include "error.h"
#include "string.h"
#include "afh.h"
#include "afs.h"
#include "list.h"
#include "ipc.h"
#include "mm.h"

/**
 * Contains statistical data of the currently admissible audio files.
 *
 * It is used to assign normalized score values to each admissible audio file.
 */
struct afs_statistics {
	/** Sum of num played over all admissible files. */
	int64_t num_played_sum;
	/** Sum of last played times over all admissible files. */
	int64_t last_played_sum;
	/** Quadratic deviation of num played time. */
	int64_t num_played_qd;
	/** Quadratic deviation of last played time. */
	int64_t last_played_qd;
	/** Number of admissible files */
	unsigned num;
};
static struct afs_statistics statistics;

/**
 * Each line of the current mood corresponds to a mood_item.
 */
struct mood_item {
	/** The method this line is referring to. */
	const struct mood_method *method;
	/** The data structure computed by the mood parser. */
	void *parser_data;
	/** The given score value, or zero if none was given. */
	int32_t score_arg;
	/** Non-zero if random scoring was requested. */
	int random_score;
	/** Whether the "not" keyword was given in the mood line. */
	int logical_not;
	/** The position in the list of items. */
	struct list_head mood_item_node;
};

/**
 * Created from the mood definition by mood_open().
 *
 * When a mood is opened, each line of its definition is investigated, and a
 * corresponding mood item is produced. Each mood line starts with \p accept,
 * \p deny, or \p score which determines the type of the mood line.  For each
 * such type a linked list is maintained whose entries are the mood items.
 *
 * \sa mood_item, mood_open().
 */
struct mood {
	/** The name of this mood. */
	char *name;
	/** The list of mood items of type \p accept. */
	struct list_head accept_list;
	/** The list of mood items of type \p deny. */
	struct list_head deny_list;
	/** The list of mood items of type \p score. */
	struct list_head score_list;
};

static struct mood *current_mood;

/**
 *  Rough approximation to sqrt.
 *
 *  \param x Integer of which to calculate the sqrt.
 *
 *  \return An integer res with res * res <= x.
 */
static uint64_t int_sqrt(uint64_t x)
{
	uint64_t op, res, one = 1;
	op = x;
	res = 0;

	one = one << 62;
	while (one > op)
		one >>= 2;

	while (one != 0) {
		if (op >= res + one) {
			op = op - (res + one);
			res = res +  2 * one;
		}
		res /= 2;
		one /= 4;
	}
//	PARA_NOTICE_LOG("sqrt(%llu) = %llu\n", x, res);
	return res;
}

/* returns 1 if row matches score item, 0 if not. */
static int get_item_score(struct mood_item *item, const struct afs_info *afsi,
		const struct afh_info *afhi, const char *path, long *score,
		long *score_arg_sum)
{
	int ret, match = 1;

	*score_arg_sum += item->random_score? 100 : PARA_ABS(item->score_arg);
	ret = 100;
	if (item->method) {
		ret = item->method->score_function(path, afsi, afhi,
			item->parser_data);
		if ((ret < 0 && !item->logical_not) || (ret >= 0 && item->logical_not))
			match = 0; /* no match */
	}
	if (item->random_score)
		*score = PARA_ABS(ret) * para_random(100);
	else
		*score = PARA_ABS(ret) * item->score_arg;
	return match;
}

/* returns 1 if row admissible, 0 if not, negative on errors */
static int compute_mood_score(const struct osl_row *aft_row, struct mood *m,
		long *result)
{
	struct mood_item *item;
	int ret, match = 0;
	long score_arg_sum = 0, score = 0, item_score;
	struct afs_info afsi;
	struct afh_info afhi;
	char *path;

	if (!m)
		return -E_NO_MOOD;
	ret = get_afsi_of_row(aft_row, &afsi);
	if (ret< 0)
		return ret;
	ret = get_afhi_of_row(aft_row, &afhi);
	if (ret< 0)
		return ret;
	ret = get_audio_file_path_of_row(aft_row, &path);
	if (ret< 0)
		return ret;
	/* reject audio file if it matches any entry in the deny list */
	list_for_each_entry(item, &m->deny_list, mood_item_node) {
		ret = get_item_score(item, &afsi, &afhi, path, &item_score,
			&score_arg_sum);
		if (ret > 0) /* not admissible */
			return 0;
		score += item_score;
	}
	list_for_each_entry(item, &m->accept_list, mood_item_node) {
		ret = get_item_score(item, &afsi, &afhi, path, &item_score,
			&score_arg_sum);
		if (ret == 0)
			continue;
		match = 1;
		score += item_score;
	}
	/* reject if there is no matching entry in the accept list */
	if (!match && !list_empty(&m->accept_list))
		return 0;
	list_for_each_entry(item, &m->score_list, mood_item_node) {
		ret = get_item_score(item, &afsi, &afhi, path, &item_score,
			&score_arg_sum);
		score += item_score;
	}
	if (score_arg_sum)
		score /= score_arg_sum;
	*result = score;
	return 1;
}

static void cleanup_list_entry(struct mood_item *item)
{
	if (item->method && item->method->cleanup)
		item->method->cleanup(item->parser_data);
	else
		free(item->parser_data);
	list_del(&item->mood_item_node);
	free(item);
}

static void destroy_mood(struct mood *m)
{
	struct mood_item *tmp, *item;

	if (!m)
		return;
	list_for_each_entry_safe(item, tmp, &m->accept_list, mood_item_node)
		cleanup_list_entry(item);
	list_for_each_entry_safe(item, tmp, &m->deny_list, mood_item_node)
		cleanup_list_entry(item);
	list_for_each_entry_safe(item, tmp, &m->score_list, mood_item_node)
		cleanup_list_entry(item);
	free(m->name);
	free(m);
}

static struct mood *alloc_new_mood(const char *name)
{
	struct mood *m = para_calloc(sizeof(struct mood));
	m->name = para_strdup(name);
	INIT_LIST_HEAD(&m->accept_list);
	INIT_LIST_HEAD(&m->deny_list);
	INIT_LIST_HEAD(&m->score_list);
	return m;
}

/** The different types of a mood line. */
enum mood_line_type {
	/** Invalid. */
	ML_INVALID,
	/** Accept line. */
	ML_ACCEPT,
	/** Deny line. */
	ML_DENY,
	/** Score line. */
	ML_SCORE
};

/** Data passed to the parser of a mood line. */
struct mood_line_parser_data {
	/** The mood this mood line belongs to. */
	struct mood *m;
	/** The line number in the mood definition. */
	unsigned line_num;
};

/*
 * <accept [with score <score>] | deny [with score <score>]  | score <score>>
 * 	[if] [not] <mood_method> [options]
 * <score> is either an integer or "random" which assigns a random score to
 * all matching files
 */

static int parse_mood_line(char *mood_line, void *data)
{
	struct mood_line_parser_data *mlpd = data;
	char **argv;
	unsigned num_words;
	char **w;
	int i, ret;
	enum mood_line_type mlt = ML_INVALID;
	struct mood_item *mi = NULL;

	mlpd->line_num++;
	ret = create_argv(mood_line, " \t", &argv);
	if (ret < 0)
		return ret;
	num_words = ret;
	if (!num_words) /* empty line */
		goto out;
	w = argv;
	if (**w == '#') /* comment */
		goto out;
	if (!strcmp(*w, "accept"))
		mlt = ML_ACCEPT;
	else if (!strcmp(*w, "deny"))
		mlt = ML_DENY;
	else if (!strcmp(*w, "score"))
		mlt = ML_SCORE;
	ret = -E_MOOD_SYNTAX;
	if (mlt == ML_INVALID)
		goto out;
	mi = para_calloc(sizeof(struct mood_item));
	if (mlt != ML_SCORE) {
		ret = -E_MOOD_SYNTAX;
		w++;
		if (!*w)
			goto out;
		if (strcmp(*w, "with"))
			goto check_for_if;
		w++;
		if (!*w)
			goto out;
		if (strcmp(*w, "score"))
			goto out;
	}
	if (mlt == ML_SCORE || !strcmp(*w, "score")) {
		ret = -E_MOOD_SYNTAX;
		w++;
		if (!*w)
			goto out;
		if (strcmp(*w, "random")) {
			mi->random_score = 0;
			ret = para_atoi32(*w, &mi->score_arg);
			if (ret < 0)
				goto out;
		} else {
			mi->random_score = 1;
			if (!*(w + 1))
			goto success; /* the line "score random" is valid */
		}
	} else
		mi->score_arg = 0;
	ret = -E_MOOD_SYNTAX;
	w++;
	if (!*w)
		goto out;
check_for_if:
	if (!strcmp(*w, "if")) {
		ret = -E_MOOD_SYNTAX;
		w++;
		if (!*w)
			goto out;
	}
	if (!strcmp(*w, "not")) {
		ret = -E_MOOD_SYNTAX;
		w++;
		if (!*w)
			goto out;
		mi->logical_not = 1;
	} else
		mi->logical_not = 0;
	for (i = 0; mood_methods[i].parser; i++) {
		if (strcmp(*w, mood_methods[i].name))
			continue;
		break;
	}
	ret = -E_MOOD_SYNTAX;
	if (!mood_methods[i].parser)
		goto out;
	ret = mood_methods[i].parser(num_words - 1 - (w - argv), w,
		&mi->parser_data);
	if (ret < 0)
		goto out;
	mi->method = &mood_methods[i];
success:
	if (mlpd->m) {
		if (mlt == ML_ACCEPT)
			para_list_add(&mi->mood_item_node, &mlpd->m->accept_list);
		else if (mlt == ML_DENY)
			para_list_add(&mi->mood_item_node, &mlpd->m->deny_list);
		else
			para_list_add(&mi->mood_item_node, &mlpd->m->score_list);
	}
	PARA_DEBUG_LOG("%s entry added, method: %p\n", mlt == ML_ACCEPT? "accept" :
		(mlt == ML_DENY? "deny" : "score"), mi->method);
	ret = 1;
out:
	free_argv(argv);
	if (ret >= 0)
		return ret;
	if (mi) {
		free(mi->parser_data);
		free(mi);
	}
	return ret;
}

static int load_mood(const struct osl_row *mood_row, struct mood **m)
{
	char *mood_name;
	struct osl_object mood_def;
	struct mood_line_parser_data mlpd = {.line_num = 0};
	int ret = mood_get_name_and_def_by_row(mood_row, &mood_name, &mood_def);

	if (ret < 0)
		return ret;
	if (!*mood_name)
		return -E_DUMMY_ROW;
	mlpd.m = alloc_new_mood(mood_name);
	ret = for_each_line_ro(mood_def.data, mood_def.size,
		parse_mood_line, &mlpd);
	osl_close_disk_object(&mood_def);
	if (ret < 0) {
		PARA_ERROR_LOG("unable to load mood %s: %s\n", mlpd.m->name,
			para_strerror(-ret));
		destroy_mood(mlpd.m);
		return ret;
	}
	*m = mlpd.m;
	return 1;
}

static int check_mood(struct osl_row *mood_row, void *data)
{
	struct para_buffer *pb = data;
	char *mood_name;
	struct osl_object mood_def;
	struct mood_line_parser_data mlpd = {.line_num = 0};

	int ret = mood_get_name_and_def_by_row(mood_row, &mood_name, &mood_def);

	if (ret < 0) {
		para_printf(pb, "failed to get mood definition: %s\n",
			para_strerror(-ret));
		return ret;
	}
	if (!*mood_name) /* ignore dummy row */
		goto out;
	ret = para_printf(pb, "checking mood %s...\n", mood_name);
	if (ret < 0)
		goto out;
	ret = for_each_line_ro(mood_def.data, mood_def.size,
		parse_mood_line, &mlpd);
	if (ret < 0)
		para_printf(pb, "%s line %u: %s\n", mood_name, mlpd.line_num,
			para_strerror(-ret));
out:
	osl_close_disk_object(&mood_def);
	return ret;
}

/**
 * Check all moods for syntax errors.
 *
 * \param fd The afs socket.
 * \param query Unused.
 */
void mood_check_callback(int fd, __a_unused const struct osl_object *query)
{
	struct para_buffer pb = {
		.max_size = SHMMAX,
		.private_data = &fd,
		.max_size_handler = pass_buffer_as_shm
	};

	int ret = para_printf(&pb, "checking moods...\n");
	if (ret < 0)
		return;
	osl_rbtree_loop(moods_table, BLOBCOL_ID, &pb,
		check_mood);
	if (pb.offset)
		pass_buffer_as_shm(pb.buf, pb.offset, &fd);
	free(pb.buf);
}

#if 0
static unsigned int_log2(uint64_t x)
{
	unsigned res = 0;

	while (x) {
		x /= 2;
		res++;
	}
	return res;
}
#endif

static int64_t normalized_value(int64_t x, int64_t n, int64_t sum, int64_t qd)
{
	if (!n || !qd)
		return 0;
	return 100 * (n * x - sum) / (int64_t)int_sqrt(n * qd);
}

static long compute_num_played_score(struct afs_info *afsi)
{
	return -normalized_value(afsi->num_played, statistics.num,
		statistics.num_played_sum, statistics.num_played_qd);
}

static long compute_last_played_score(struct afs_info *afsi)
{
	return -normalized_value(afsi->last_played, statistics.num,
		statistics.last_played_sum, statistics.last_played_qd);
}

static long compute_dynamic_score(const struct osl_row *aft_row)
{
	struct afs_info afsi;
	int64_t score, nscore = 0, lscore = 0;
	int ret;

	ret = get_afsi_of_row(aft_row, &afsi);
	if (ret < 0)
		return -100;
	nscore = compute_num_played_score(&afsi);
	lscore = compute_last_played_score(&afsi);
	score = nscore + lscore;
	return score;
}

static int add_afs_statistics(const struct osl_row *row)
{
	uint64_t n, x, s;
	struct afs_info afsi;
	int ret;

	ret = get_afsi_of_row(row, &afsi);
	if (ret < 0)
		return ret;
	n = statistics.num;
	x = afsi.last_played;
	s = statistics.last_played_sum;
	if (n > 0)
		statistics.last_played_qd += (x - s / n) * (x - s / n) * n / (n + 1);
	statistics.last_played_sum += x;

	x = afsi.num_played;
	s = statistics.num_played_sum;
	if (n > 0)
		statistics.num_played_qd += (x - s / n) * (x - s / n) * n / (n + 1);
	statistics.num_played_sum += x;
	statistics.num++;
	return 1;
}

static int del_afs_statistics(const struct osl_row *row)
{
	uint64_t n, s, q, a, new_s;
	struct afs_info afsi;
	int ret;
	ret = get_afsi_of_row(row, &afsi);
	if (ret < 0)
		return ret;
	n = statistics.num;
	assert(n);
	if (n == 1) {
		memset(&statistics, 0, sizeof(statistics));
		return 1;
	}

	s = statistics.last_played_sum;
	q = statistics.last_played_qd;
	a = afsi.last_played;
	new_s = s - a;
	statistics.last_played_sum = new_s;
	statistics.last_played_qd = q + s * s / n - a * a
		- new_s * new_s / (n - 1);

	s = statistics.num_played_sum;
	q = statistics.num_played_qd;
	a = afsi.num_played;
	new_s = s - a;
	statistics.num_played_sum = new_s;
	statistics.num_played_qd = q + s * s / n - a * a
		- new_s * new_s / (n - 1);

	statistics.num--;
	return 1;
}

/**
 * Structure used during mood_open().
 *
 * At mood open time, we look at each file in the audio file table in order to
 * determine whether it is admissible. If a file happens to be admissible, its
 * mood score is computed by calling each relevant mood_score_function. Next,
 * we update the afs_statistics and add a struct admissible_file_info to a
 * temporary array.
 *
 * If all files have been processed that way, the final score of each
 * admissible file is computed by adding the dynamic score (which depends on
 * the afs_statistics) to the mood score.  Finally, all audio files in the
 * array are added to the score table and the admissible array is freed.
 *
 * \sa mood_method, admissible_array.
 */
struct admissible_file_info
{
	/** The admissible audio file. */
	struct osl_row *aft_row;
	/** Its score. */
	long score;
};

/** The temporary array of admissible files. */
struct admissible_array {
	/** Files are admissible wrt. this mood. */
	struct mood *m;
	/** The size of the array */
	unsigned size;
	/** Pointer to the array of admissible files. */
	struct admissible_file_info *array;
};

/**
 * Add an entry to the array of admissible files.
 *
 * \param aft_row The audio file to be added.
 * \param private_data Pointer to a struct admissible_file_info.
 *
 * \return 1 if row admissible, 0 if not, negative on errors.
 */
static int add_if_admissible(struct osl_row *aft_row, void *data)
{
	struct admissible_array *aa = data;
	int ret;
	long score = 0;

	ret = compute_mood_score(aft_row, aa->m, &score);
	if (ret <= 0)
		return ret;
	if (statistics.num >= aa->size) {
		aa->size *= 2;
		aa->size += 100;
		aa->array = para_realloc(aa->array,
			aa->size * sizeof(struct admissible_file_info));
	}
	aa->array[statistics.num].aft_row = aft_row;
	aa->array[statistics.num].score = score;
	ret = add_afs_statistics(aft_row);
	if (ret < 0)
		return ret;
	return 1;
}

/**
 * Compute the new quadratic deviation in case one element changes.
 *
 * \param n Number of elements.
 * \param old_qd The quadratic deviation before the change.
 * \param old_val The value that was replaced.
 * \param new_val The replacement value.
 * \param old_sum The sum of all elements before the update.
 *
 * \return The new quadratic deviation resulting from replacing old_val
 * by new_val.
 *
 * Given n real numbers a_1, ..., a_n, their sum S = a_1 + ... + a_n,
 * their quadratic deviation
 *
 * q = (a_1 - S/n)^2 + ... + (a_n - S/n)^2,
 *
 * and a real number b, the quadratic deviation q' of a_1,...a_{n-1}, b (ie.
 * the last number a_n was replaced by b) may be computed in O(1) time in terms
 * of n, q, a_n, b, and S as
 *
 * 	q' = q + d * s - (2 * S + d) * d / n,
 *
 * where d = b - a_n, and s = b + a_n.
 *
 * Example: n = 3, a_1 = 3, a_2 = 5, a_3 = 7, b = 10. Then S = 15, q = 8, d = 3,
 * s = 17, so
 *
 * 	q + d * s - (2 * S + d) * d / n = 8 + 51 - 33 = 26,
 *
 * which equals q' = (3 - 6)^2 + (5 - 6)^2 + (10 - 6)^2.
 *
 */
_static_inline_ int64_t update_quadratic_deviation(int64_t n, int64_t old_qd,
		int64_t old_val, int64_t new_val, int64_t old_sum)
{
	int64_t delta = new_val - old_val;
	int64_t sigma = new_val + old_val;
	return old_qd + delta * sigma - (2 * old_sum + delta) * delta / n;
}

static int update_afs_statistics(struct afs_info *old_afsi, struct afs_info *new_afsi)
{
	unsigned n;
	int ret = get_num_admissible_files(&n);

	if (ret < 0)
		return ret;
	assert(n);

	statistics.last_played_qd = update_quadratic_deviation(n,
		statistics.last_played_qd, old_afsi->last_played,
		new_afsi->last_played, statistics.last_played_sum);
	statistics.last_played_sum += new_afsi->last_played - old_afsi->last_played;

	statistics.num_played_qd = update_quadratic_deviation(n,
		statistics.num_played_qd, old_afsi->num_played,
		new_afsi->num_played, statistics.num_played_sum);
	statistics.num_played_sum += new_afsi->num_played - old_afsi->num_played;
	return 1;
}

static int add_to_score_table(const struct osl_row *aft_row, long mood_score)
{
	long score = (compute_dynamic_score(aft_row) + mood_score) / 3;
	return score_add(aft_row, score);
}

static int delete_from_statistics_and_score_table(const struct osl_row *aft_row)
{
	int ret = del_afs_statistics(aft_row);
	if (ret < 0)
		return ret;
	return score_delete(aft_row);
}

/**
 * Delete one entry from the statistics and from the score table.
 *
 * \param aft_row The audio file which is no longer admissible.
 *
 * \return Positive on success, negative on errors.
 *
 * \sa score_delete().
 */
static int mood_delete_audio_file(const struct osl_row *aft_row)
{
	int ret;

	ret = row_belongs_to_score_table(aft_row, NULL);
	if (ret < 0)
		return ret;
	if (!ret) /* not admissible, nothing to do */
		return 1;
	return delete_from_statistics_and_score_table(aft_row);
}

/**
 * Compute the new score of an audio file wrt. the current mood.
 *
 * \param aft_row Determines the audio file.
 * \param old_afsi The audio file selector info before updating.
 *
 * The \a old_afsi argument may be \p NULL which indicates that no changes to
 * the audio file info were made.
 *
 * \return Positive on success, negative on errors.
 */
static int mood_update_audio_file(const struct osl_row *aft_row,
		struct afs_info *old_afsi)
{
	long score, percent;
	int ret, is_admissible, was_admissible = 0;
	struct afs_info afsi;
	unsigned rank;

	if (!current_mood)
		return 1; /* nothing to do */
	ret = row_belongs_to_score_table(aft_row, &rank);
	if (ret < 0)
		return ret;
	was_admissible = ret;
	ret = compute_mood_score(aft_row, current_mood, &score);
	if (ret < 0)
		return ret;
	is_admissible = (ret > 0);
	if (!was_admissible && !is_admissible)
		return 1;
	if (was_admissible && !is_admissible)
		return delete_from_statistics_and_score_table(aft_row);
	if (!was_admissible && is_admissible) {
		ret = add_afs_statistics(aft_row);
		if (ret < 0)
			return ret;
		return add_to_score_table(aft_row, score);
	}
	/* update score */
	ret = get_afsi_of_row(aft_row, &afsi);
	if (ret < 0)
		return ret;
	if (old_afsi) {
		ret = update_afs_statistics(old_afsi, &afsi);
		if (ret < 0)
			return ret;
	}
	score += compute_num_played_score(&afsi);
	score += compute_last_played_score(&afsi);
	score /= 3;
	PARA_DEBUG_LOG("score: %li\n", score);
	percent = (score + 100) / 3;
	if (percent > 100)
		percent = 100;
	else if (percent < 0)
		percent = 0;
	PARA_DEBUG_LOG("moving from rank %u to %lu%%\n", rank, percent);
	return score_update(aft_row, percent);
}

static void log_statistics(void)
{
	unsigned n = statistics.num;

	if (!n) {
		PARA_NOTICE_LOG("no admissible files\n");
		return;
	}
	PARA_INFO_LOG("last_played mean: %lli, last_played sigma: %llu\n",
		(long long int)(statistics.last_played_sum / n),
		(long long unsigned)int_sqrt(statistics.last_played_qd / n));
	PARA_INFO_LOG("num_played mean: %lli, num_played sigma: %llu\n",
		(long long int)statistics.num_played_sum / n,
		(long long unsigned)int_sqrt(statistics.num_played_qd / n));
}

/**
 * Close the current mood.
 *
 * Free all resources of the current mood which were allocated during
 * mood_open().
 */
void close_current_mood(void)
{
	destroy_mood(current_mood);
	current_mood = NULL;
	memset(&statistics, 0, sizeof(statistics));
}

/**
 * Change the current mood.
 *
 * \param mood_name The name of the mood to open.
 *
 * If \a mood_name is \a NULL, load the dummy mood that accepts every audio file
 * and uses a scoring method based only on the \a last_played information.
 *
 * If there is already an open mood, it will be closed first.
 *
 * \return Positive on success, negative on errors. Loading the dummy mood
 * always succeeds.
 *
 * \sa struct admissible_file_info, struct admissible_array, struct
 * afs_info::last_played, mood_close().
 */
int change_current_mood(char *mood_name)
{
	int i, ret;
	struct admissible_array aa = {
		.size = 0,
		.array = NULL
	};

	if (mood_name) {
		struct mood *m;
		struct osl_row *row;
		struct osl_object obj = {
			.data = mood_name,
			.size = strlen(mood_name) + 1
		};
		ret = osl(osl_get_row(moods_table, BLOBCOL_NAME, &obj, &row));
		if (ret < 0) {
			PARA_NOTICE_LOG("no such mood: %s\n", mood_name);
			return ret;
		}
		ret = load_mood(row, &m);
		if (ret < 0)
			return ret;
		close_current_mood();
		current_mood = m;
	} else {
		close_current_mood();
		current_mood = alloc_new_mood("dummy");
	}
	aa.m = current_mood;
	PARA_NOTICE_LOG("computing statistics of admissible files\n");
	ret = audio_file_loop(&aa, add_if_admissible);
	if (ret < 0)
		return ret;
	log_statistics();
	PARA_INFO_LOG("%d admissible files \n", statistics.num);
	for (i = 0; i < statistics.num; i++) {
		struct admissible_file_info *a = aa.array + i;
		ret = add_to_score_table(a->aft_row, a->score);
		if (ret < 0)
			goto out;
	}
	PARA_NOTICE_LOG("loaded mood %s\n", current_mood->name);
	ret = statistics.num;
out:
	free(aa.array);
	return ret;
}
/**
 * Close and re-open the current mood.
 *
 * This function is used if changes to the audio file table or the
 * attribute table were made that render the current list of admissible
 * files useless. For example, if an attribute is removed from the
 * attribute table, this function is called.
 *
 * \return Positive on success, negative on errors. If no mood is currently
 * open, the function returns success.
 *
 * \sa mood_open(), mood_close().
 */
static int reload_current_mood(void)
{
	int ret;
	char *mood_name = NULL;

	if (!current_mood)
		return 1;
	PARA_NOTICE_LOG("reloading %s\n", current_mood->name?
		current_mood->name : "(dummy)");
	if (current_mood->name)
		mood_name = para_strdup(current_mood->name);
	close_current_mood();
	ret = change_current_mood(mood_name);
	free(mood_name);
	return ret;
}

/**
 * Notification callback for the moods table.
 *
 * \param event Type of the event just occurred.
 * \param pb Unused.
 * \param data Its type depends on the event.
 *
 * This function performs actions required due to the occurrence of the given
 * event. Possible actions include reload of the current mood and update of the
 * score of an audio file.
 */
int moods_event_handler(enum afs_events event, __a_unused struct para_buffer *pb,
		void *data)
{
	if (!current_mood)
		return 0;
	switch (event) {
	/*
	 * The three blob events might change the set of admissible files,
	 * so we must reload the score list.
	 */
	case BLOB_RENAME:
	case BLOB_REMOVE:
	case BLOB_ADD:
		if (data == moods_table || data == playlists_table)
			return 1; /* no reload necessary for these */
		return reload_current_mood();
	/* these also require reload of the score table */
	case ATTRIBUTE_ADD:
	case ATTRIBUTE_REMOVE:
	case ATTRIBUTE_RENAME:
		return reload_current_mood();
	/* changes to the aft only require to re-examine the audio file */
	case AFSI_CHANGE: {
		struct afsi_change_event_data *aced = data;
		return mood_update_audio_file(aced->aft_row, aced->old_afsi);
		}
	case AFHI_CHANGE:
	case AUDIO_FILE_RENAME:
	case AUDIO_FILE_ADD:
		return mood_update_audio_file(data, NULL);
	case AUDIO_FILE_REMOVE:
		return mood_delete_audio_file(data);
	default:
		return 1;
	}
}
